#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for the QKD routing protocol.
"""

from typing import List
from qcompute_qnet.core.des import EventHandler
from qcompute_qnet.protocols.routing import Routing
from qcompute_qnet.models.qkd.key_generation import PrepareAndMeasure
from qcompute_qnet.models.qkd.message import QKDMessage
from qcompute_qnet.models.qkd.resource_management import QKDRMP

__all__ = [
    "QKDRouting",
    "QKDRoutingL2"
]


class QKDRouting(Routing):
    r"""Class for the routing protocol in quantum key distribution.

    Attributes:
        path (List[Node]): path of key swapping
        upstream_node (Node): upstream node in the path
        downstream_node (Node): downstream node in the path
        keys (Dict[Node, List]): keys generated with upstream and downstream nodes
    """

    def __init__(self, name: str):
        r"""Constructor for QKDRouting class.

        Args:
            name (str): name of the QKDRouting protocol
        """
        super().__init__(name)
        self.path = None
        self.upstream_node = None
        self.downstream_node = None
        self.keys = {}

    def receive_upper(self, upper_protocol: type, **kwargs) -> None:
        r"""Receive a message from an upper protocol.

        Args:
            upper_protocol (type): upper protocol that sends the message
            **kwargs: received keyword arguments
        """
        msg = kwargs['msg']

        # Alice: Send a 'REQUEST' message
        if msg.data['type'] == QKDMessage.Type.REQUEST:
            # Check if the destination is directly connected to the node
            if msg.dst in self.node.direct_nodes:
                self.node.send_classical_msg(dst=msg.dst, msg=msg)
            else:
                self.node.send_classical_msg(dst=self.node.direct_repeater, msg=msg)

        # Bob: Send back an 'ACCEPT' message
        elif msg.data['type'] == QKDMessage.Type.ACCEPT:
            self.path = msg.data['path']  # save the path for key swapping
            self.node.env.logger.info(f"Path of the QKD request: {[node.name for node in self.path]}")

            self.upstream_node = self.path[-2]  # get the upstream node from the path
            key_num = msg.data['key_num']
            key_length = msg.data['key_length']

            self.node.send_classical_msg(dst=self.upstream_node, msg=msg)  # forward the 'ACCEPT' to the upstream node

            # Create an instance for key generation with the upstream node
            self.node.set_key_generation(self.upstream_node)
            self.send_lower(PrepareAndMeasure,
                            peer=self.upstream_node,
                            role=PrepareAndMeasure.Role.RECEIVER,
                            key_num=key_num,
                            key_length=key_length)

        # Bob: Send back the 'ACKNOWLEDGE' message to the repeater node
        elif msg.data['type'] == QKDMessage.Type.ACKNOWLEDGE:
            if msg.dst in self.node.direct_nodes:
                self.node.send_classical_msg(dst=msg.dst, msg=msg)
            else:
                self.node.send_classical_msg(dst=self.node.direct_repeater, msg=msg)

        # Bob: Send back the 'DONE' message to Alice
        elif msg.data['type'] == QKDMessage.Type.DONE:
            if msg.dst in self.node.direct_nodes:
                self.node.send_classical_msg(dst=msg.dst, msg=msg)
            else:
                self.node.send_classical_msg(dst=self.node.direct_repeater, msg=msg)

            self.reset()

    def receive_lower(self, lower_protocol: type, **kwargs) -> None:
        r"""Receive a message from a lower protocol.

        Args:
            lower_protocol (type): lower protocol that sends the message
            **kwargs: received keyword arguments
        """
        from qcompute_qnet.models.qkd.node import TrustedRepeaterNode, EndNode
        from qcompute_qnet.models.qkd.application import QKDApp

        peer = kwargs['peer']
        self.keys[peer] = kwargs['sifted_keys']

        # EndNode: Deliver the sifted keys to the upper protocol
        if isinstance(self.node, EndNode):
            # If the two users are directly connected, then finish the QKD
            finish = True if len(self.path) == 2 else False
            ready_msg = QKDMessage(src=None,
                                   dst=None,
                                   protocol=QKDApp,
                                   data={'type': QKDMessage.Type.READY,
                                         'sifted_keys': self.keys[peer],
                                         'finish': finish})
            self.send_upper(QKDApp, msg=ready_msg)

        # Trusted repeater node: Do key swapping
        elif isinstance(self.node, TrustedRepeaterNode):
            # Check if already generated keys for both directions
            if self.upstream_node in self.keys.keys() and self.downstream_node in self.keys.keys():
                ciphertext_list = []
                # Do 'xor' operation to each pair of keys to generate a ciphertext list
                for key_down, key_up in zip(self.keys[self.downstream_node], self.keys[self.upstream_node]):
                    ciphertext_list.append(key_down ^ key_up)

                # Send the ciphertext list to Bob
                cipher_msg = QKDMessage(src=self.node,
                                        dst=self.path[-1],
                                        protocol=QKDRouting,
                                        data={'type': QKDMessage.Type.CIPHERTEXT,
                                              'ciphertext_list': ciphertext_list})
                next_hop = self.node.classical_routing_table[cipher_msg.dst]
                self.node.send_classical_msg(dst=next_hop, msg=cipher_msg)

    def receive_classical_msg(self, msg: "QKDMessage", **kwargs) -> None:
        r"""Receive a QKDMessage from the node.

        Args:
            msg (QKDMessage): received QKDMessage
            **kwargs (Any): keyword arguments for receiving the classical message
        """
        from qcompute_qnet.models.qkd.node import EndNode, TrustedRepeaterNode
        from qcompute_qnet.models.qkd.application import QKDApp

        if msg.data['type'] == QKDMessage.Type.REQUEST:
            self.node.env.logger.info(f"{self.node.name} received 'REQUEST' of QKDApp "
                                      f"with {msg.dst.name} from {msg.src.name}")

            # Repeater: Receive the 'REQUEST' from Alice
            if isinstance(self.node, TrustedRepeaterNode):
                msg.data['path'].append(self.node)  # append 'self.node' to the path

                # Find the next hop node and forward the 'REQUEST' message
                next_hop = self.node.quantum_routing_table[msg.dst]
                self.node.send_classical_msg(dst=next_hop, msg=msg)

            # Bob: Receive the 'REQUEST' from Alice
            elif isinstance(self.node, EndNode):
                # Create a QKDApp instance and send the request
                self.node.set_qkd_app()
                self.send_upper(QKDApp, msg=msg)

        elif msg.data['type'] == QKDMessage.Type.ACCEPT:
            self.node.env.logger.info(f"{self.node.name} received 'ACCEPT' of QKDApp from {msg.src.name}")

            self.path = msg.data['path']  # save the path for key swapping
            key_num = msg.data['key_num']
            key_length = msg.data['key_length']

            # Repeater: Receive the 'ACCEPT' message
            if isinstance(self.node, TrustedRepeaterNode):
                # Get upstream node and downstream node according to the path
                index = msg.data['path'].index(self.node)
                self.upstream_node, self.downstream_node = msg.data['path'][index - 1], msg.data['path'][index + 1]

                self.node.send_classical_msg(self.upstream_node, msg=msg)  # forward 'ACCEPT' to the upstream node

                # Set lower 'PrepareAndMeasure' protocols for key generation with the upstream node and downstream node
                self.node.set_key_generation(self.upstream_node)
                self.send_lower(PrepareAndMeasure,
                                peer=self.upstream_node,
                                role=PrepareAndMeasure.Role.RECEIVER,
                                key_num=key_num,
                                key_length=key_length)

                self.node.set_key_generation(self.downstream_node)
                self.send_lower(PrepareAndMeasure,
                                peer=self.downstream_node,
                                role=PrepareAndMeasure.Role.TRANSMITTER,
                                key_num=key_num,
                                key_length=key_length)

            # Alice: Receive the 'ACCEPT' message from Bob
            elif isinstance(self.node, EndNode):
                self.downstream_node = self.path[1]  # get downstream node from the path

                # Begin key generation with downstream node
                self.node.set_key_generation(self.downstream_node)
                self.send_lower(PrepareAndMeasure,
                                peer=self.downstream_node,
                                role=PrepareAndMeasure.Role.TRANSMITTER,
                                key_num=key_num,
                                key_length=key_length)

        elif msg.data['type'] == QKDMessage.Type.CIPHERTEXT:
            # Bob: Receive ciphertext list from a repeater node
            if msg.dst == self.node:
                self.node.env.logger.info(f"{self.node.name} received 'CIPHERTEXT' of QKDApp from {msg.src.name}")
                self.send_upper(QKDApp, msg=msg)

            # Repeater: Receive the 'CIPHERTEXT' from other repeaters, directly forward the message to the destination
            else:
                next_hop = self.node.classical_routing_table[msg.dst]
                self.node.send_classical_msg(dst=next_hop, msg=msg)

        # Repeater: Receive the 'ACKNOWLEDGE' from Bob
        elif msg.data['type'] == QKDMessage.Type.ACKNOWLEDGE:
            if msg.dst == self.node:
                self.node.env.logger.info(f"{self.node.name} received 'ACKNOWLEDGE' for ciphertext of QKDApp"
                                          f" from {msg.src.name}")
                self.reset()  # reset the QKD routing protocol

            else:
                next_hop = self.node.classical_routing_table[msg.dst]
                self.node.send_classical_msg(dst=next_hop, msg=msg)

        elif msg.data['type'] == QKDMessage.Type.DONE:
            # Alice: Receive the 'DONE' from Bob
            if msg.dst == self.node:
                self.node.env.logger.info(f"{self.node.name} received 'DONE' of QKDApp from {msg.src.name}")
                handler = EventHandler(self, "send_upper", [QKDApp], msg=msg)
                self.node.scheduler.schedule_now(handler)
                self.reset()  # reset the QKD routing protocol

            # Repeater: Directly forward the 'DONE' message
            else:
                next_hop = self.node.classical_routing_table[msg.dst]
                self.node.send_classical_msg(dst=next_hop, msg=msg)

    def reset(self) -> None:
        r"""Reset the QKD routing protocol.
        """
        for proto in self.lower_protocols:
            if isinstance(proto, PrepareAndMeasure) and \
                    (proto.peer == self.upstream_node or proto.peer == self.downstream_node):
                # Remove the PrepareAndMeasure protocol instances from the protocol stack
                self.owner.stack.remove_node(proto)
                proto.owner = proto
        self.path = None
        self.upstream_node = None
        self.downstream_node = None
        self.keys = {}


class QKDRoutingL2(QKDRouting):
    r"""Class for the routing protocol in resource management architecture for quantum key distribution.

    Attributes:
        keys_up (Dict): save keys with upstream nodes
        keys_down (Dict): save keys with downstream nodes
        serving_req_id (str): identifier of the serving request
    """

    def __init__(self, name: str):
        r"""Constructor for QKDRoutingL2 class.

        Args:
            name (str): name of the QKDRoutingL2 protocol
        """
        super().__init__(name)
        self.keys_up = {}
        self.keys_down = {}
        self.serving_req_id = None

    def receive_upper(self, upper_protocol: type, **kwargs) -> None:
        r"""Receive a message from an upper protocol.

        Args:
            upper_protocol (type): upper protocol that sends the message
            **kwargs: received keyword arguments
        """
        from qcompute_qnet.models.qkd.node import RMPEndNode, RMPRepeaterNode

        msg = kwargs['msg']

        if msg.data['type'] == QKDMessage.Type.REQUEST:
            if isinstance(self.node, RMPEndNode):
                # Check if the destination is directly connected to the node
                if msg.dst in self.node.direct_nodes:
                    self.node.send_classical_msg(dst=msg.dst, msg=msg)
                elif isinstance(self.node.direct_repeater, RMPRepeaterNode):
                    self.node.send_classical_msg(dst=self.node.direct_repeater, msg=msg)
                else:
                    self.node.env.logger.info(f"Illegal key request: "
                                              f"no RMP repeater between {msg.src.name} and {msg.dst.name}")

                    # Send a 'REJECT' to QKDRMP protocol
                    reject_msg = QKDMessage(src=msg.src,
                                            dst=msg.dst,
                                            protocol=QKDRMP,
                                            data={'type': QKDMessage.Type.REJECT,
                                                  'req_id': msg.data['req_id'],
                                                  'path': msg.data['path'],
                                                  'err_type': 1,
                                                  'key_num': msg.data['key_num'],
                                                  'key_length': msg.data['key_length']})
                    self.send_upper(upper_protocol=QKDRMP, msg=reject_msg)

            elif isinstance(self.node, RMPRepeaterNode):
                next_hop = self.node.quantum_routing_table[msg.dst]
                self.node.send_classical_msg(dst=next_hop, msg=msg)

        elif msg.data['type'] == QKDMessage.Type.ACCEPT:
            option = kwargs['option']

            # Condition 1: Forward the message
            if option == 'forward':
                # Forward the 'ACCEPT' to the upstream node
                self.path = msg.data['path']
                index = msg.data['path'].index(self.node)
                upstream_node = msg.data['path'][index - 1]
                self.node.send_classical_msg(upstream_node, msg=msg)

                if self.node == msg.src:
                    key_num = msg.data['key_num']
                    key_length = msg.data['key_length']
                    # Set the lower protocol for key generation with its upstream node
                    self.node.set_key_generation(upstream_node)
                    self.send_lower(PrepareAndMeasure,
                                    peer=upstream_node,
                                    role=PrepareAndMeasure.Role.RECEIVER,
                                    key_num=key_num,
                                    key_length=key_length)

            # Condition 2: Process the request
            elif option == 'operate':
                pattern = kwargs['pattern']

                # Pattern 0: RMPRepeaterNode on both sides
                if pattern == 0:
                    self.path = msg.data['path']
                    index = self.path.index(self.node)
                    upstream_node, downstream_node = msg.data['path'][index - 1], msg.data['path'][index + 1]

                    key_num = msg.data['key_num']
                    key_length = msg.data['key_length']

                    # Fetch sifted keys from key pools at both side and generate the ciphertext
                    ciphertext_list = []
                    keys_up = self.node.key_pools[upstream_node][1].allocate(key_num, key_length, towards='up')
                    keys_down = self.node.key_pools[downstream_node][0].allocate(key_num, key_length, towards='down')

                    # Record the fetched keys
                    self.node.env.logger.info(f"{self.node.name} fetched the keys with upstream node "
                                              f"{upstream_node.name} from the key pool: {keys_up}")
                    self.node.env.logger.info(f"{self.node.name} fetched the keys with downstream node "
                                              f"{downstream_node.name} from the key pool: {keys_down}")

                    for key_up, key_down in zip(keys_up, keys_down):
                        ciphertext_list.append(key_up ^ key_down)

                    # Send the ciphertext list to the receiver
                    cipher_msg = QKDMessage(src=self.node,
                                            dst=self.path[-1],
                                            protocol=QKDRoutingL2,
                                            data={'type': QKDMessage.Type.CIPHERTEXT,
                                                  'req_id': msg.data['req_id'],
                                                  'path': msg.data['path'],
                                                  'ciphertext_list': ciphertext_list,
                                                  'key_num': msg.data['key_num'],
                                                  'key_length': msg.data['key_length']})
                    self.node.send_classical_msg(dst=downstream_node, msg=cipher_msg)

                    # Record the key delivery
                    self.node.reqs_delivered += 1
                    self.node.keys_delivered += key_num

                    self.serving_req_id = None

                    # Call the QKDRMP to pop a request
                    pop_msg = QKDMessage(src=self.node,
                                         dst=self.node,
                                         protocol=QKDRMP,
                                         data={'type': QKDMessage.Type.POP})
                    self.send_upper(QKDRMP, msg=pop_msg)

                # Pattern 1: RMPEndNode on one side and RMPRepeaterNode on the other side
                elif pattern == 1:
                    self.serving_req_id = msg.data['req_id']
                    self.path = msg.data['path']
                    index = self.path.index(self.node)
                    upstream_node, downstream_node = msg.data['path'][index - 1], msg.data['path'][index + 1]

                    key_num = msg.data['key_num']
                    key_length = msg.data['key_length']
                    towards = kwargs['towards']

                    if towards == 'up':
                        key_pool_index = 1
                        repeater_peer = upstream_node
                        end_peer = downstream_node
                        key_end_dict = self.keys_up

                    elif towards == 'down':
                        key_pool_index = 0
                        repeater_peer = downstream_node
                        end_peer = upstream_node
                        key_end_dict = self.keys_down

                    # Fetch the keys from the corresponding key pool
                    ciphertext_list = []
                    keys_repeater = self.node.key_pools[repeater_peer][key_pool_index]. \
                        allocate(key_num, key_length, towards=towards)
                    self.node.env.logger.info(f"{self.node.name} fetched the keys with {repeater_peer.name}"
                                              f" from the key pool: {keys_repeater}")

                    # Check if the key generation on the other side is finished
                    if key_end_dict[end_peer]['keys'] is None:
                        # Save the delivered keys and wait
                        if key_pool_index == 0:
                            self.keys_up[tuple([repeater_peer, end_peer])] = \
                                {'path': self.path, 'keys': keys_repeater}
                        else:
                            self.keys_down[tuple([repeater_peer, end_peer])] = \
                                {'path': self.path, 'keys': keys_repeater}

                    else:
                        # Calculate the ciphertext
                        for key_repeater, key_end in zip(keys_repeater, key_end_dict[end_peer]['keys']):
                            ciphertext_list.append(key_repeater ^ key_end)
                        # Clear the key dict
                        key_end_dict.pop(end_peer)

                        # Send the ciphertext list to the receiver
                        cipher_msg = QKDMessage(src=self.node,
                                                dst=self.path[-1],
                                                protocol=QKDRoutingL2,
                                                data={'type': QKDMessage.Type.CIPHERTEXT,
                                                      'req_id': msg.data['req_id'],
                                                      'path': msg.data['path'],
                                                      'ciphertext_list': ciphertext_list,
                                                      'key_num': msg.data['key_num'],
                                                      'key_length': msg.data['key_length']})
                        self.node.send_classical_msg(dst=downstream_node, msg=cipher_msg)
                        self.node.env.logger.info(f"{self.node.name} generated the ciphertext with {repeater_peer.name}"
                                                  f" and {end_peer.name}, and sent it to {cipher_msg.dst.name}")

                        # Record the key delivery
                        self.node.reqs_delivered += 1
                        self.node.keys_delivered += key_num

                        self.serving_req_id = None

                        # Ask QKDRMP to pop a request
                        pop_msg = QKDMessage(src=self.node,
                                             dst=self.node,
                                             protocol=QKDRMP,
                                             data={'type': QKDMessage.Type.POP})
                        self.send_upper(QKDRMP, msg=pop_msg)

        elif msg.data['type'] == QKDMessage.Type.REJECT:
            err_type = msg.data['err_type']

            if err_type == 1:
                # Forward the 'REJECT' to the upstream node
                upstream_node = msg.data['path'][-2]
                self.node.send_classical_msg(upstream_node, msg=msg)

            elif err_type == 2:
                # Forward the 'REJECT' to the upstream node
                self.path = msg.data['path']
                index = msg.data['path'].index(self.node)
                upstream_node = msg.data['path'][index - 1]

                self.node.send_classical_msg(upstream_node, msg=msg)

            elif err_type == 3:
                # Forward the 'REJECT' to the downstream nodes to cancel the resource reservation
                self.path = msg.data['path']
                index = msg.data['path'].index(self.node)
                downstream_node = msg.data['path'][index + 1]

                self.node.send_classical_msg(downstream_node, msg=msg)

        elif msg.data['type'] == QKDMessage.Type.CONFIRM:
            # Forward the 'CONFIRM' to the downstream node
            self.path = msg.data['path']
            index = msg.data['path'].index(self.node)
            downstream_node = msg.data['path'][index + 1]

            self.node.send_classical_msg(downstream_node, msg=msg)

        elif msg.data['type'] == QKDMessage.Type.ACKNOWLEDGE:
            # Send the message to the next hop
            upstream_node = msg.data['path'][-2]
            self.node.send_classical_msg(dst=upstream_node, msg=msg)

        elif msg.data['type'] == QKDMessage.Type.DONE:
            # Send 'DONE' to the upstream node
            upstream_node = msg.data['path'][-2]
            self.node.send_classical_msg(upstream_node, msg=msg)

        elif msg.data['type'] == QKDMessage.Type.GENERATE:
            towards = kwargs['towards']
            if isinstance(msg.dst, RMPEndNode):
                if towards == 'up':
                    self.keys_up[msg.dst] = {'path': msg.data['path'], 'keys': None}
                elif towards == 'down':
                    self.keys_down[msg.dst] = {'path': msg.data['path'], 'keys': None}

            # Call the key generation protocol to start key generation
            self.node.set_key_generation(msg.dst, towards=towards)
            role = PrepareAndMeasure.Role.TRANSMITTER if towards == 'up' else PrepareAndMeasure.Role.RECEIVER
            self.send_lower(PrepareAndMeasure,
                            peer=msg.dst,
                            role=role,
                            key_num=msg.data['key_num'],
                            key_length=msg.data['key_length'])

    def receive_lower(self, lower_protocol: type, **kwargs) -> None:
        r"""Receive a message from a lower protocol.

        Args:
          lower_protocol (type): lower protocol that sends the message
          **kwargs: received keyword arguments
        """
        from qcompute_qnet.models.qkd.node import RMPRepeaterNode, RMPEndNode

        peer = kwargs['peer']
        self.keys[peer] = kwargs['sifted_keys']
        towards = kwargs['towards']

        if isinstance(self.node, RMPRepeaterNode):
            if isinstance(peer, RMPRepeaterNode):
                # Store the generated keys in the corresponding key pool
                index = 0 if towards == 'up' else 1
                for key in self.keys[peer]:
                    self.node.key_pools[peer][index].keys.append(key)
                    self.node.key_pools[peer][index].current_volume += 1
                # Record the key delivery
                self.node.env.logger.info(f"The sifted keys between {self.node.name} and {peer.name} at "
                                          f"{self.node.name}'s key pools: {self.node.key_pools[peer][index].keys}")
                # Reset the status of the corresponding key pool
                self.node.key_pools[peer][index].generating = False
                self.node.key_pools[peer][index].idle = True

                # Check if all the key pools of the node are ready for delivering keys
                ready_to_deliver = True
                for _, key_pools in self.node.key_pools.items():
                    for key_pool in key_pools:
                        if key_pool.generating is True:
                            ready_to_deliver = False
                            break

                if ready_to_deliver is True:
                    # Call QKDRMP to pop a request
                    self.node.env.logger.info(f"Key pools at {self.node.name} were all ready to allocate keys,"
                                              f" asked the request queue of RMP protocol to pop a request")

                    pop_msg = QKDMessage(src=self.node,
                                         dst=self.node,
                                         protocol=QKDRMP,
                                         data={'type': QKDMessage.Type.POP})
                    self.send_upper(QKDRMP, msg=pop_msg)

            elif isinstance(peer, RMPEndNode):
                # Check if keys at another side is delivered
                if towards == 'up':
                    key_end_dict = self.keys_up
                    key_repeater_dict = self.keys_down
                elif towards == 'down':
                    key_end_dict = self.keys_down
                    key_repeater_dict = self.keys_up
                # Get the repeater_peer
                if peer == key_end_dict[peer]['path'][0]:
                    repeater_peer = key_end_dict[peer]['path'][2]
                elif peer == key_end_dict[peer]['path'][-1]:
                    repeater_peer = key_end_dict[peer]['path'][-3]

                if isinstance(repeater_peer, RMPRepeaterNode):
                    if key_repeater_dict.get(tuple([repeater_peer, peer])) is not None:
                        cipher_text_list = []
                        # Fetch keys from the corresponding key pool
                        for key_repeater, key_end in \
                                zip(self.keys[peer], key_repeater_dict[tuple([repeater_peer, peer])]['keys']):
                            cipher_text_list.append(key_repeater ^ key_end)

                        key_repeater_dict.pop(tuple([repeater_peer, peer]))
                        self.keys.pop(peer)
                        self.path = key_end_dict[peer]['path']
                        key_end_dict.pop(peer)

                        # Generate the 'CIPHERTEXT' message
                        cipher_msg = QKDMessage(src=self.node,
                                                dst=self.path[-1],
                                                protocol=QKDRoutingL2,
                                                data={'type': QKDMessage.Type.CIPHERTEXT,
                                                      'req_id': self.serving_req_id,
                                                      'path': self.path,
                                                      'ciphertext_list': cipher_text_list,
                                                      'key_num': len(cipher_text_list),
                                                      'key_length': len(str(cipher_text_list[0]))})
                        # Record the delivery
                        self.node.reqs_delivered += 1
                        self.node.keys_delivered += cipher_msg.data['key_num']
                        self.serving_req_id = None
                        # Record the delivery to log file
                        self.node.env.logger.info(f"{self.node.name} fetched the allocated keys,"
                                                  f" generated the ciphertext with {peer.name} and"
                                                  f" {repeater_peer.name}, sent it to {cipher_msg.dst.name}")
                        # Send the ciphertext to the receiver
                        downstream_node_index = self.path.index(self.node) + 1
                        downstream_node = self.path[downstream_node_index]
                        self.node.send_classical_msg(dst=downstream_node, msg=cipher_msg)

                        # Call QKDRMP to pop a request
                        pop_msg = QKDMessage(src=self.node,
                                             dst=self.node,
                                             protocol=QKDRMP,
                                             data={'type': QKDMessage.Type.POP})
                        self.send_upper(QKDRMP, msg=pop_msg)

                    else:
                        # Save the keys to the corresponding key pool
                        if towards == 'up':
                            self.keys_up[peer]['keys'] = self.keys[peer]
                        else:
                            self.keys_down[peer]['keys'] = self.keys[peer]
                        self.node.env.logger.info(f"Key generation at {self.node.name} with {peer.name} was ready,"
                                                  f" waiting the key pool at another side to get ready")

                elif isinstance(repeater_peer, RMPEndNode):
                    if key_repeater_dict[repeater_peer]['keys'] is not None:
                        cipher_text_list = []
                        # Fetch keys from the corresponding keypool
                        for key_repeater, key_end in zip(self.keys[peer],
                                                         key_repeater_dict[repeater_peer]['keys']):
                            cipher_text_list.append(key_repeater ^ key_end)

                        # Clear the two key dicts
                        key_repeater_dict.pop(repeater_peer)
                        self.keys.pop(peer)
                        self.path = key_end_dict[peer]['path']
                        key_end_dict.pop(peer)

                        # Generate the 'CIPHERTEXT' message
                        cipher_msg = QKDMessage(src=self.node,
                                                dst=self.path[-1],
                                                protocol=QKDRoutingL2,
                                                data={'type': QKDMessage.Type.CIPHERTEXT,
                                                      'req_id': self.serving_req_id,
                                                      'path': self.path,
                                                      'ciphertext_list': cipher_text_list,
                                                      'key_num': len(cipher_text_list),
                                                      'key_length': len(str(cipher_text_list[0]))})
                        # Record the delivery
                        self.node.reqs_delivered += 1
                        self.node.keys_delivered += cipher_msg.data['key_num']
                        self.serving_req_id = None
                        # Record the event to log file
                        self.node.env.logger.info(f"{self.node.name} fetched the allocated keys,"
                                                  f" generated the ciphertext with {peer.name} and"
                                                  f" {repeater_peer.name}, sent it to {cipher_msg.dst.name}")
                        # Send the ciphertext to the receiver
                        downstream_node_index = self.path.index(self.node) + 1
                        downstream_node = self.path[downstream_node_index]
                        self.node.send_classical_msg(dst=downstream_node, msg=cipher_msg)

                        # Call QKDRMP to pop a request
                        pop_msg = QKDMessage(src=self.node,
                                             dst=self.node,
                                             protocol=QKDRMP,
                                             data={'type': QKDMessage.Type.POP})
                        self.send_upper(QKDRMP, msg=pop_msg)

                    else:
                        # Save the keys to corresponding key pool
                        if towards == 'up':
                            self.keys_up[peer]['keys'] = self.keys[peer]
                        else:
                            self.keys_down[peer]['keys'] = self.keys[peer]
                        # Record the waiting to log file
                        self.node.env.logger.info(f"Key generation at {self.node.name} with {peer.name} was ready, waiting"
                                                  f" for the key pool at another side to get ready")

        elif isinstance(self.node, RMPEndNode):
            # If the two users are directly connected, end the QKD request
            finish = True if len(self.path) == 2 else False

            self.node.env.logger.info(f"{self.node.name} finished the key generation with {peer.name}")

            # Generate a 'READY' message and send it to the QKDRMP
            ready_msg = QKDMessage(src=None,
                                   dst=None,
                                   protocol=QKDRMP,
                                   data={'type': QKDMessage.Type.READY,
                                         'sifted_keys': self.keys[peer],
                                         'finish': finish})
            self.send_upper(QKDRMP, msg=ready_msg, peer=peer, towards=towards)

    def receive_classical_msg(self, msg: "QKDMessage", **kwargs) -> None:
        r"""Receive QKDMessage from other nodes.

        Args:
            msg (QKDMessage): received QKDMessage
            **kwargs : keyword arguments for receiving the classical message
        """
        from qcompute_qnet.models.qkd.node import RMPEndNode, RMPRepeaterNode

        if msg.data['type'] == QKDMessage.Type.REQUEST:
            self.node.env.logger.info(f"{self.node.name} received 'REQUEST' of QKD "
                                      f"with {msg.dst.name} from {msg.src.name}")

            # Generate a 'PATH' message and send it to QKDRMP
            request_msg = QKDMessage(src=msg.src,
                                     dst=msg.dst,
                                     protocol=QKDRMP,
                                     data={'type': QKDMessage.Type.PATH,
                                           'req_id': msg.data['req_id'],
                                           'path': msg.data['path'],
                                           'key_num': msg.data['key_num'],
                                           'key_length': msg.data['key_length']})
            self.send_upper(QKDRMP, msg=request_msg)

        elif msg.data['type'] == QKDMessage.Type.ACCEPT:
            self.node.env.logger.info(f"{self.node.name} received 'ACCEPT' of QKD "
                                      f"with {msg.dst.name} from {msg.src.name}")

            if isinstance(self.node, RMPEndNode):
                # Sender: receive 'ACCEPT' message from the receiver
                self.path = msg.data['path']
                downstream_node = self.path[1]
                key_num = msg.data['key_num']
                key_length = msg.data['key_length']

                # Create an instance for key generation with the upstream node
                self.node.set_key_generation(downstream_node)
                self.send_lower(PrepareAndMeasure,
                                peer=downstream_node,
                                role=PrepareAndMeasure.Role.TRANSMITTER,
                                key_num=key_num,
                                key_length=key_length)

            # Generate a 'RESERVE' message and send it to QKDRMP
            reserve_msg = QKDMessage(src=msg.src,
                                     dst=msg.dst,
                                     protocol=QKDRMP,
                                     data={'type': QKDMessage.Type.RESERVE,
                                           'req_id': msg.data['req_id'],
                                           'path': msg.data['path'],
                                           'key_num': msg.data['key_num'],
                                           'key_length': msg.data['key_length']})
            self.send_upper(QKDRMP, msg=reserve_msg)

        elif msg.data['type'] == QKDMessage.Type.CONFIRM:
            if self.node == msg.dst:
                self.send_upper(QKDRMP, msg=msg)

            else:
                # Forward the 'CONFIRM' to the next hop
                self.path = msg.data['path']
                index = self.path.index(self.node)
                self.downstream_node = msg.data['path'][index + 1]
                self.node.send_classical_msg(dst=self.downstream_node, msg=msg)

        elif msg.data['type'] == QKDMessage.Type.REJECT:
            self.node.env.logger.info(f"{self.node.name} received reject from {msg.src.name} to {msg.dst.name}")

            err_type = msg.data['err_type']
            if err_type == 3:
                # Generate a 'REJECT' message and send to QKDRMP
                discard_msg = QKDMessage(src=msg.src,
                                         dst=msg.dst,
                                         protocol=QKDRMP,
                                         data={'type': QKDMessage.Type.REJECT,
                                               'req_id': msg.data['req_id'],
                                               'path': msg.data['path'],
                                               'err_type': msg.data['err_type'],
                                               'key_num': msg.data['key_num'],
                                               'key_length': msg.data['key_length']})
                self.send_upper(QKDRMP, msg=discard_msg)

                # For repeater nodes, forward the message to the downstream node
                if self.node != msg.dst:
                    self.path = msg.data['path']
                    index = self.path.index(self.node)
                    self.downstream_node = msg.data['path'][index + 1]
                    self.node.send_classical_msg(dst=self.downstream_node, msg=msg)

            else:
                if self.node == msg.dst:
                    # Generate a 'REJECT' message and send to QKDRMP
                    discard_msg = QKDMessage(src=msg.src,
                                             dst=msg.dst,
                                             protocol=QKDRMP,
                                             data={'type': QKDMessage.Type.REJECT,
                                                   'req_id': msg.data['req_id'],
                                                   'path': msg.data['path'],
                                                   'err_type': msg.data['err_type'],
                                                   'key_num': msg.data['key_num'],
                                                   'key_length': msg.data['key_length']})
                    self.send_upper(QKDRMP, msg=discard_msg)

                else:
                    # Forward the message to downstream node
                    self.path = msg.data['path']
                    index = self.path.index(self.node)
                    upstream_node = msg.data['path'][index - 1]
                    self.node.send_classical_msg(dst=upstream_node, msg=msg)

        elif msg.data['type'] == QKDMessage.Type.CIPHERTEXT:
            if msg.dst == self.node:
                self.node.env.logger.info(f"{self.node.name} received ciphertexts from {msg.src.name}")

                # Generate a 'CIPHERTEXT' message and send it to QKDRMP
                cipher_msg = QKDMessage(src=msg.src,
                                        dst=msg.dst,
                                        protocol=QKDRMP,
                                        data={'type': QKDMessage.Type.CIPHERTEXT,
                                              'req_id': msg.data['req_id'],
                                              'path': msg.data['path'],
                                              'ciphertext_list': msg.data['ciphertext_list'],
                                              'key_num': msg.data['key_num'],
                                              'key_length': msg.data['key_length']})
                self.send_upper(QKDRMP, msg=cipher_msg)

            else:
                # Forward the message to the next hop
                if isinstance(self.node, RMPRepeaterNode):
                    self.node.env.logger.info(f"{self.node.name} received ciphertext"
                                              f" from {msg.src.name} to {msg.dst.name}")

                    path = msg.data['path']
                    index = path.index(self.node)
                    next_hop = path[index + 1]
                    self.node.send_classical_msg(dst=next_hop, msg=msg)

        elif msg.data['type'] == QKDMessage.Type.ACKNOWLEDGE:
            if msg.dst == self.node:
                # Send 'DONE' to QKDRMP
                ack_msg = QKDMessage(src=msg.src,
                                     dst=msg.dst,
                                     protocol=QKDRMP,
                                     data={'type': QKDMessage.Type.ACKNOWLEDGE,
                                           'req_id': msg.data['req_id'],
                                           'path': msg.data['path'],
                                           'key_num': msg.data['key_num'],
                                           'key_length': msg.data['key_length']})
                self.send_upper(QKDRMP, msg=ack_msg)

            else:
                # Forward the message to the upstream node
                index = msg.data['path'].index(self.node)
                next_hop = msg.data['path'][index - 1]
                self.node.send_classical_msg(next_hop, msg=msg)

        elif msg.data['type'] == QKDMessage.Type.GENERATE:
            towards = kwargs['towards']
            role = PrepareAndMeasure.Role.TRANSMITTER if towards == 'up' else PrepareAndMeasure.Role.RECEIVER

            # Get ready for key generation
            self.node.set_key_generation(msg.dst, towards=towards)
            self.send_lower(PrepareAndMeasure,
                            peer=msg.dst,
                            role=role,
                            key_num=msg.data['key_num'],
                            key_length=msg.data['key_length'])

            if isinstance(msg.dst, RMPEndNode):
                if towards == 'up':
                    self.keys_up[msg.dst] = {'path': msg.data['path'], 'keys': None}
                elif towards == 'down':
                    self.keys_down[msg.dst] = {'path': msg.data['path'], 'keys': None}

        elif msg.data['type'] == QKDMessage.Type.DONE:
            if msg.dst == self.node:
                # Send a 'DONE' message to QKDRMP protocol
                done_msg = QKDMessage(src=msg.src,
                                      dst=msg.dst,
                                      protocol=QKDRMP,
                                      data={'type': QKDMessage.Type.DONE,
                                            'req_id': msg.data['req_id'],
                                            'path': msg.data['path'],
                                            'key_num': msg.data['key_num'],
                                            'key_length': msg.data['key_length']})
                self.send_upper(QKDRMP, msg=done_msg)

            else:
                # Forward the 'DONE' to the next hop
                if self.node in msg.data['path']:
                    index = msg.data['path'].index(self.node)
                    next_hop = msg.data['path'][index - 1]
                    self.node.send_classical_msg(dst=next_hop, msg=msg)
                else:
                    self.node.env.logger.info(f"wrong node {self.node.name} receive a DONE from {msg.src.name} to {msg.dst.name}")

    def send_lower(self, lower_protocol: type, **kwargs) -> None:
        r"""Send message to lower protocol.

        Args:
            lower_protocol (type): type of the lower protocol to receive the message
            **kwargs: keyword arguments of the message
        """
        assert not self.is_bottom, f"'{self.name}' has no lower protocols."

        for proto in self.lower_protocols:
            if isinstance(proto, lower_protocol) and proto.role is None:
                proto.receive_upper(type(self), **kwargs)
                break



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for communication channels.
"""

from abc import abstractmethod, ABC
from numpy import random
from qcompute_qnet.core.des import Entity, EventHandler

__all__ = [
    "Channel",
    "ClassicalChannel",
    "QuantumChannel",
    "ClassicalFiberChannel",
    "QuantumFiberChannel",
    "DuplexChannel",
    "DuplexClassicalFiberChannel",
    "DuplexQuantumFiberChannel",
    "ClassicalFreeSpaceChannel",
    "QuantumFreeSpaceChannel"
]


LIGHT_SPEED = 3e-4  # meters per picoseconds


class Channel(Entity, ABC):
    r"""Class for the simulation of communication channels.

    Attributes:
        sender (Node): sender of the channel
        receiver (Node): receiver of the channel
        delay (int): delay of message transmission in picoseconds
        loss (float): loss of the channel in decibels
        lossy_prob (float): lossy probability of the channel
    """

    def __init__(self, name: str, distance: float, loss=0, env=None, sender=None, receiver=None, delay=None):
        r"""Constructor for Channel class.

        Args:
            name (str): name of the channel
            distance (float): length of the channel in meters
            loss (float): loss of the channel in decibels
            env (DESEnv): discrete-event simulation environment
            sender (Node): sender of the channel
            receiver (Node): receiver of the channel
            delay (int): delay of message transmission in picoseconds
        """
        super().__init__(name, env)
        self.__distance = distance
        self.delay = round(distance / LIGHT_SPEED) if delay is None else delay
        self.loss = loss
        self.lossy_prob = 0

        self.sender, self.receiver = None, None
        if sender is not None and receiver is not None:
            self.connect(sender, receiver)

    def init(self) -> None:
        r"""Channel initialization.

        This method will do a sanity check to confirm that the channel is installed to a ``Link``.
        """
        assert self.owner != self, f"The channel {self.name} should be installed to a 'Link' first!"

    def set_distance(self, distance: float) -> None:
        r"""Set a distance for the channel.

        Args:
            distance (float): distance to set
        """
        self.__distance = distance

    def get_distance(self) -> float:
        r"""Get the distance of the channel.

        Returns:
            float: distance of the channel
        """
        return self.__distance

    def set_loss(self, loss: float) -> None:
        r"""Set a loss for the channel.

        Args:
            loss (float): loss to set
        """
        self.loss = loss

    def connect(self, sender: "Node", receiver: "Node") -> None:
        r"""Connect a sender node and a receiver node.

        Args:
            sender (Node): sender node
            receiver (Node): receiver node
        """
        assert self.sender is None and self.receiver is None, \
            f"Channel has been connected to '{self.sender.name}' and '{self.receiver.name}'."
        self.sender, self.receiver = sender, receiver

    @abstractmethod
    def transmit(self, msg: "Message", priority=None) -> None:
        r"""Abstract method for message transmission of the channel.

        This method should be implemented by a specific channel.

        Args:
            msg (Message): message to transmit
            priority (int): priority of the transmission
        """
        pass


class ClassicalChannel(Channel):
    r"""Class for the simulation of classical channels.

    Note:
        ``ClassicalChannel`` is a logical classical channel in which no transmission loss will take place.
         Thus, the lossy probability of a classical channel is set to be zero.

    Examples:
        We can create a classical channel between two nodes (e.g. alice and bob) by two approaches.
        The first approach is to instantiate a ``ClassicalChannel`` with given ``name`` and ``distance``.
        Then using the ``connect`` method to make a connection to alice and bob.

        >>> c_a2b = ClassicalChannel(name="c_a2b", distance=1e3)
        >>> c_a2b.connect(alice, bob)

        The second approach is to assign the sender and receiver when instantiating.
        This will make your code more compact.

        >>> c_a2b = ClassicalChannel(name="c_a2b", distance=1e3, sender=alice, receiver=bob)
    """

    def __init__(self, name: str, distance: float, loss=0, env=None, sender=None, receiver=None, delay=None):
        r"""Constructor for ClassicalChannel class.

        Args:
            name (str): name of the classical channel
            distance (float): length of the classical channel in meters
            loss (float): loss of the classical channel in decibels
            env (DESEnv): discrete-event simulation environment
            sender (Node): sender of the classical channel
            receiver (Node): receiver of the classical channel
            delay (int): delay of message transmission in picoseconds
        """
        super().__init__(name, distance, loss, env, sender, receiver, delay)

    def transmit(self, msg: "ClassicalMessage", priority=None) -> None:
        r"""Transmit classical message from the sender to the receiver.

        Args:
            msg (ClassicalMessage): classical message to transmit
            priority (int): priority of the transmission event
        """
        handler = EventHandler(self.receiver, "receive_classical_msg", [self.sender, msg])
        self.scheduler.schedule_after(self.delay, handler, priority)


class QuantumChannel(Channel):
    r"""Class for the simulation of quantum channels.

    Note:
        ``QuantumChannel`` is a logical quantum channel in which no transmission loss will take place.
         Thus, the lossy probability of a quantum channel is set to be zero.

    Examples:
        We can create a quantum channel between two nodes (e.g. alice and bob) by two approaches.
        The first approach is to instantiate a ``QuantumChannel`` with given ``name`` and ``distance``.
        Then using the ``connect`` method to make a connection to alice and bob.

        >>> q_a2b = QuantumChannel(name="q_a2b", distance=1e3)
        >>> q_a2b.connect(alice, bob)

        The second approach is to assign the sender and receiver when instantiating.
        This will make your code more compact.

        >>> q_a2b = QuantumChannel(name="q_a2b", distance=1e3, sender=alice, receiver=bob)
    """

    def __init__(self, name: str, distance: float, loss=0, env=None, sender=None, receiver=None, delay=None):
        r"""Constructor for QuantumChannel class.

        Args:
            name (str): name of the quantum channel
            distance (float): length of the quantum channel in meters
            loss (float): loss of the quantum channel in decibels
            env (DESEnv): discrete-event simulation environment
            sender (Node): sender of the quantum channel
            receiver (Node): receiver of the quantum channel
            delay (int): delay of message transmission in picoseconds
        """
        super().__init__(name, distance, loss, env, sender, receiver, delay)

    def transmit(self, msg: "QuantumMessage", priority=None) -> None:
        r"""Transmit quantum message from the sender to the receiver.

        Args:
            msg (QuantumMessage): quantum message to transmit
            priority (int): priority of the transmission event
        """
        handler = EventHandler(self.receiver, "receive_quantum_msg", [self.sender, msg])
        self.scheduler.schedule_after(self.delay, handler, priority)


class ClassicalFiberChannel(ClassicalChannel):
    r"""Class for the simulation of classical fiber channels.

    Note:
        The lossy probability of a classical fiber channel is set to be zero for simplicity.
    """

    def __init__(self, name: str, distance: float, loss=0, env=None, sender=None, receiver=None, delay=None):
        r"""Constructor for ClassicalFiberChannel class.

        Args:
            name (str): name of the classical fiber channel
            distance (float): length of the classical fiber channel in meters
            loss (float): loss of the classical fiber channel in decibels
            env (DESEnv): discrete-event simulation environment
            sender (Node): sender of the classical fiber channel
            receiver (Node): receiver of the classical fiber channel
            delay (int): delay of message transmission in picoseconds
        """
        super().__init__(name, distance, loss, env, sender, receiver, delay)


class QuantumFiberChannel(QuantumChannel):
    r"""Class for the simulation of quantum fiber channels.

    Note:
        We use the standard commercial single mode fiber (SMF) to model the channel loss.
        The lossy probability is given by :math:`1 - 10^{- loss / 10}`, where loss (dB) is calculated by
        :math:`\alpha * L`, :math:`\alpha` is the attenuation (dB/km, default 0.2) and
        :math:`L` is the channel length (km).
    """

    def __init__(self, name: str, distance: float, loss=None, env=None, sender=None, receiver=None, delay=None):
        r"""Constructor for QuantumFiberChannel class.

        Args:
            name (str): name of the quantum fiber channel
            distance (float): length of the quantum fiber channel in meters
            loss (float): loss of the quantum fiber channel in decibels
            env (DESEnv): discrete-event simulation environment
            sender (Node): sender of the quantum fiber channel
            receiver (Node): receiver of the quantum fiber channel
            delay (int): delay of message transmission in picoseconds
            loss (float): loss of the quantum fiber channel
        """
        super().__init__(name, distance, loss, env, sender, receiver, delay)
        self.loss = 0.2 * distance * 1e-3 if loss is None else loss  # set default attenuation as 0.2
        self.lossy_prob = 1 - 10 ** (- self.loss / 10)

    def transmit(self, msg: "QuantumMessage", priority=None) -> None:
        r"""Transmit quantum message from the sender to the receiver.

        Args:
            msg (QuantumMessage): quantum message to transmit
            priority (int): priority of the transmission event
        """
        # Schedule an event for the receiver if the message is not lost
        if random.random_sample() > self.lossy_prob:
            super().transmit(msg, priority)


class DuplexChannel(Entity, ABC):
    r"""Class for the simulation of duplex channels.

    A duplex channel contains two unidirectional channels. These two unidirectional channels are identical except for
    their directions. One must set the ends of two channels at the same time by 'connect' method.

    Attributes:
        node1 (Node): one end of the duplex channel
        node2 (Node): the other end of the duplex channel
        channel1_2 (Channel): the channel from node1 to node2
        channel2_1 (Channel): the channel from node2 to node1
        delay (int): delay of message transmission in picoseconds
        lossy_prob (float): lossy probability of the duplex channel
    """

    def __init__(self, name: str, distance: float, loss=0, env=None, node1=None, node2=None, delay=None):
        r"""Constructor for DuplexChannel class.

        Args:
            name (str): name of the duplex channel
            distance (float): length of the duplex channel in meters
            loss (float): loss of the duplex channel in decibels
            env (DESEnv): discrete-event simulation environment
            node1 (Node): one end of the duplex channel
            node2 (Node): the other end of the duplex channel
            delay (int): delay of message transmission in picoseconds
        """
        super().__init__(name, env)
        self.node1 = node1
        self.node2 = node2
        self.channel1_2 = None
        self.channel2_1 = None
        self.__distance = distance
        self.loss = loss
        self.delay = delay
        self.lossy_prob = 0

    def init(self) -> None:
        r"""Duplex channel initialization.

        This method will do a sanity check to confirm that the duplex channel is installed to a ``Link``.
        """
        assert self.owner != self, f"The duplex channel {self.name} should be installed to a 'Link' first!"

    def set_distance(self, distance: float) -> None:
        r"""Set a distance for the two unidirectional channels of the duplex channel.

        Args:
            distance (float): distance to set
        """
        self.channel1_2.set_distance(distance)
        self.channel2_1.set_distance(distance)

    def get_distance(self) -> float:
        r"""Get the distance of the duplex channel.

        Returns:
            float: distance of the channel
        """
        return self.__distance

    def set_loss(self, loss: float) -> None:
        r"""Set a loss for the two unidirectional channels of the duplex channel.

        Args:
            loss (float): loss to set
        """
        self.channel1_2.set_loss(loss)
        self.channel2_1.set_loss(loss)

    def connect(self, node1: "Node", node2: "Node") -> None:
        r"""Connect a sender node and a receiver node for both unidirectional channels of the duplex channel.

        Args:
            node1 (Node): one end of the duplex channel
            node2 (Node): the other end of the duplex channel
        """
        self.channel1_2.connect(node1, node2)
        self.channel1_2.name = node1.name + "->" + node2.name
        self.channel2_1.connect(node2, node1)
        self.channel2_1.name = node2.name + "->" + node1.name


class DuplexClassicalFiberChannel(DuplexChannel):
    r"""Class for the simulation of duplex classical fiber channels.

    A duplex classical fiber channel contains two unidirectional classical fiber channels. These two unidirectional
    channels are identical except for their directions. One must set the ends of two channels at the same time by
    'connect' method.
    """

    def __init__(self, name: str, distance: float, loss=0, env=None, node1=None, node2=None, delay=None):
        r"""Constructor for DuplexClassicalFiberChannel class.

        Args:
            name (str): name of the duplex classical fiber channel
            distance (float): length of the duplex classical fiber channel in meters
            loss (float): loss of the duplex classical fiber channel in decibels
            env (DESEnv): discrete-event simulation environment
            node1 (Node): one end of the duplex classical fiber channel
            node2 (Node): the other end of the duplex classical fiber channel
            delay (int): delay of message transmission in picoseconds
        """
        super().__init__(name, distance, loss, env, node1, node2, delay)
        self.channel1_2 = ClassicalFiberChannel("node1->node2", distance, loss, env=env,
                                                sender=node1, receiver=node2, delay=delay)
        self.channel2_1 = ClassicalFiberChannel("node2->node1", distance, loss, env=env,
                                                sender=node2, receiver=node1, delay=delay)
        self.install([self.channel1_2, self.channel2_1])


class DuplexQuantumFiberChannel(DuplexChannel):
    r"""Class for the simulation of duplex quantum fiber channels.

    A duplex quantum fiber channel contains two unidirectional quantum fiber channels. These two unidirectional
    channels are identical except for their directions. One must set the ends of two channels at the same time by
    'connect' method.
    """

    def __init__(self, name: str, distance: float, loss=None, env=None, node1=None, node2=None, delay=None):
        r"""Constructor for DuplexQuantumFiberChannel class.

        Args:
            name (str): name of the duplex quantum fiber channel
            distance (float): length of the duplex quantum fiber channel in meters
            loss (float): loss of the duplex quantum fiber channel in decibels
            env (DESEnv): discrete-event simulation environment
            node1 (Node): one end of the duplex quantum fiber channel
            node2 (Node): the other end of the duplex quantum fiber channel
            delay (int): delay of message transmission in picoseconds
        """
        super().__init__(name, distance, loss, env, node1, node2, delay)
        self.channel1_2 = QuantumFiberChannel("node1->node2", distance, loss, env=env,
                                              sender=node1, receiver=node2, delay=delay)
        self.channel2_1 = QuantumFiberChannel("node2->node1", distance, loss, env=env,
                                              sender=node2, receiver=node1, delay=delay)
        self.install([self.channel1_2, self.channel2_1])
        self.loss = 0.2 * distance * 1e-3 if loss is None else loss  # set default attenuation as 0.2
        self.lossy_prob = 1 - 10 ** (- self.loss / 10)


class ClassicalFreeSpaceChannel(ClassicalChannel):
    r"""Class for the simulation of a free space channel for transmitting classical message.

    Warning:
        1. If there is a mobile node, the value of ``distance`` is calculated according to the track of its mobile node.
        2. If both of the connected nodes are fixed, we should set a value for ``distance``.
    """

    def __init__(self, name: str, distance=0, loss=0, is_mobile=False,
                 env=None, sender=None, receiver=None, delay=None):
        r"""Constructor for ClassicalFreeSpaceChannel class.

        Args:
            name (str): name of the classical free space channel
            distance (float): length of the classical free space channel in meters
            loss (float): loss of the classical free space channel in decibels per kilometer
            is_mobile (bool): whether the classical free space channel is mobile
            env (DESEnv): discrete-event simulation environment
            sender (Node): sender of the classical free space channel
            receiver (Node): receiver of the classical free space channel
            delay (int): delay of message transmission in picoseconds
        """
        assert is_mobile or distance != 0, "Should set a distance if both connected nodes are fixed!"

        self.is_mobile = is_mobile
        self.mobile_node = None
        super().__init__(name, distance, loss, env, sender, receiver, delay)

    def connect(self, sender: "Node", receiver: "Node") -> None:
        r"""Connect a sender node and a receiver node.

        Args:
            sender (Node): sender node
            receiver (Node): receiver node
        """
        super().connect(sender, receiver)
        if self.is_mobile:
            assert not (sender.is_mobile and receiver.is_mobile), \
                f"We do not allow both nodes to be mobile in this version."
            if sender.is_mobile:
                self.mobile_node = sender
            elif receiver.is_mobile:
                self.mobile_node = receiver
            assert self.mobile_node is not None, "The channel is claimed to be mobile but both nodes are fixed."

            self.set_distance(self.mobile_node.mobility.track.time2distance(self.mobile_node.env.now))
            self.delay = round(self.get_distance() / LIGHT_SPEED)

    def update_params(self, distance: float) -> None:
        r"""Update the parameters of the classical free space channel.

        Args:
            distance (float): updated distance of the quantum free space channel
        """
        self.set_distance(distance)
        # Update transmission delay of the classical message
        self.delay = round(self.get_distance() / LIGHT_SPEED)

    def transmit(self, msg: "ClassicalMessage", priority=None) -> None:
        r"""Transmit classical message from the sender to the receiver.

        Args:
            msg (ClassicalMessage): classical message to transmit
            priority (int): priority of the transmission event
        """
        # Acquire the current distance between the two connected nodes
        distance = self.mobile_node.mobility.track.time2distance(self.env.now)
        self.update_params(distance)

        handler = EventHandler(self.receiver, "receive_classical_msg", [self.sender, msg])
        self.scheduler.schedule_after(self.delay, handler, priority)


class QuantumFreeSpaceChannel(QuantumChannel):
    r"""Class for the simulation of a free space channel for transmitting quantum message.

    Warning:
        1. If there is a mobile node, the value of ``distance`` is calculated according to the track of its mobile node.
        2. If both of the connected nodes are fixed, we should set a value for ``distance``.
    """

    def __init__(self, name: str, distance=0, loss=None, is_mobile=False,
                 env=None, sender=None, receiver=None, delay=None):
        r"""Constructor for QuantumFreeSpaceChannel class.

        Args:
            name (str): name of the quantum free space channel
            distance (float): length of the quantum free space channel in meters
            loss (float): loss of the quantum free space channel in decibels per kilometer
            is_mobile (bool): whether the quantum free space channel is mobile
            env (DESEnv): discrete-event simulation environment
            sender (Node): sender of the quantum free space channel
            receiver (Node): receiver of the quantum free space channel
            delay (int): delay of message transmission in picoseconds
            loss (float): loss of the quantum free space channel
        """
        assert is_mobile or distance != 0, "Should set a distance if both connected nodes are fixed!"
        assert is_mobile or loss is not None, "Should set a loss if both connected nodes are fixed!"

        self.is_mobile = is_mobile
        self.mobile_node = None
        super().__init__(name, distance, loss, env, sender, receiver, delay)
        self.lossy_prob = 1 - 10 ** (- self.loss / 10)

    def connect(self, sender: "Node", receiver: "Node") -> None:
        r"""Connect a sender node and a receiver node.

        Args:
            sender (Node): sender node
            receiver (Node): receiver node
        """
        super().connect(sender, receiver)
        if self.is_mobile:
            assert not (sender.is_mobile and receiver.is_mobile), \
                f"We do not allow both nodes to be mobile in this version."
            if sender.is_mobile:
                self.mobile_node = sender
            elif receiver.is_mobile:
                self.mobile_node = receiver
            assert self.mobile_node is not None, "The channel is claimed to be mobile but both nodes are fixed."

            self.set_distance(self.mobile_node.mobility.track.time2distance(self.mobile_node.env.now))
            self.delay = round(self.get_distance() / LIGHT_SPEED)
            self.set_loss(self.mobile_node.mobility.track.distance2loss(self.get_distance()))

    def update_params(self, distance: float, loss: float) -> None:
        r"""Update the parameters of the quantum free space channel.

        Args:
            distance (float): updated distance of the quantum free space channel
            loss (float): updated loss of the quantum free space channel
        """
        self.set_distance(distance)
        # Update transmission delay of the quantum message
        self.delay = round(self.get_distance() / LIGHT_SPEED)
        # Update loss and lossy probability of the quantum message
        self.set_loss(loss)
        self.lossy_prob = 1 - 10 ** (- self.loss / 10)

    def transmit(self, msg: "QuantumMessage", priority=None) -> None:
        r"""Transmit quantum message from the sender to the receiver.

        Args:
            msg (QuantumMessage): quantum message to transmit
            priority (int): priority of the transmission event
        """
        # Acquire the current distance and loss of the channel
        distance = self.mobile_node.mobility.track.time2distance(self.env.now)
        loss = self.mobile_node.mobility.track.distance2loss(distance)
        self.update_params(distance, loss)

        # Schedule an event for the receiver if the message is not lost
        if random.random_sample() > self.lossy_prob:
            handler = EventHandler(self.receiver, "receive_quantum_msg", [self.sender, msg])
            self.scheduler.schedule_after(self.delay, handler, priority)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for protocol and request queue in resource management.
"""

from abc import ABC
from typing import List
from qcompute_qnet.protocols.protocol import Protocol

__all__ = [
    "RMP",
    "RequestQueue"
]


class RMP(Protocol, ABC):
    r"""Abstract protocol class for the resource management protocol.

    Args:
        name (str): name of the protocol
    """

    def __init__(self, name: str):
        r"""Constructor of the resource management protocol.

        Args:
             name (str): name of the resource management protocol
        """
        super().__init__(name)


class RequestQueue:
    r"""Class for the request queue in resource management protocol.

    Attributes:
        max_volume (int): maximum number of requests that the request queue can accommodate
        current_volume (int): current number of requests in the request queue
        requests (List[ClassicalMessage]): the list of the requests
        node (Node): the node that holds the request queue
    """

    def __init__(self, node: "Node", max_volume: int):
        r"""Constructor for RequestQueue class.

        Args:
            node (Node): the node that holds the request queue
            max_volume (int): maximum number of requests that the request queue can accommodate
        """
        self.node = node
        self.max_volume = max_volume
        self.current_volume = 0
        self.requests = []

    def push(self, request: "ClassicalMessage") -> bool:
        r"""Push a request into the request queue.

        Args:
            request (ClassicalMessage): the request to be pushed

        Returns:
            bool: whether the request is pushed into the request queue successfully
        """
        if self.current_volume != self.max_volume:
            self.requests.insert(0, request)
            self.current_volume += 1
            return True
        else:
            return False

    def pop(self, strategy: str = '_fcfs') -> "ClassicalMessage":
        r"""Pop a request from the request queue according to some certain strategies.
        The first-come-first-served algorithm is adopted as the default strategy.

        Args:
            strategy (str): the strategy for popping the request

        Returns:
            ClassicalMessage: the request to execute
        """
        assert self.current_volume != 0, f"There's no request in the request queue of {self.node.name}"

        # Pop a request according to a certain strategy
        func = getattr(self, strategy)
        self.current_volume -= 1
        return func()

    def _fcfs(self) -> "ClassicalMessage":
        r"""First-come-first-served strategy.

        Returns:
            ClassicalMessage: popped request selected by the first-come-first-served strategy
        """
        return self.requests.pop()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for messages.
"""

from abc import ABC
from typing import Any, Dict

__all__ = [
    "Message",
    "ClassicalMessage",
    "QuantumMessage",
]


class Message(ABC):
    r"""Abstract class for the classical message in a network.

    Attributes:
        data: message content
    """

    def __init__(self, data: Any):
        r"""Constructor for Message class.

        Args:
            data (Any): message content
        """
        self.data = data


class ClassicalMessage(Message):
    r"""Class for the classical message as the carrier of classical information.

    Attributes:
        src (Node): source of the classical message
        dst (Node): destination of the classical message
        protocol (type) : protocol related to the classical message
        data (Dict): message content
    """

    def __init__(self, src: "Node", dst: "Node", protocol: type, data: Dict):
        r"""Constructor for ClassicalMessage class.

        Args:
            src (Node): source of the classical message
            dst (Node): destination of the classical message
            protocol (type) : protocol related to the classical message
            data (Dict): message content
        """
        super().__init__(data)
        self.src = src
        self.dst = dst
        self.protocol = protocol


class QuantumMessage(Message):
    r"""Class for the quantum message as the carrier of quantum information.

    Attributes:
        data (QuantumState): quantum state of the quantum message
    """

    def __init__(self, data: "QuantumState"):
        r"""Constructor for QuantumMessage class.

        Args:
            data (QuantumState): quantum state of the quantum message
        """
        super().__init__(data)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Install library to site-packages.
"""

from __future__ import absolute_import
try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='qcompute-qnet',
    version='1.2.0',
    description='A Quantum NETwork toolkit developed by the Institute for Quantum Computing at Baidu Research.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Institute for Quantum Computing, Baidu INC.',
    author_email='quantum@baidu.com',
    url="https://quantum-hub.baidu.com/qnet/tutorial-introduction",
    python_requires=">=3.8",
    packages=find_packages(),
    install_requires=[
        'qcompute==3.1.0',
        'numpy>=1.19.3',
        'pandas>=1.4.2',
        'networkx>=2.8',
        'matplotlib>=3.5.1'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    license='Apache 2.0',
)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for basic quantum network protocols.
"""

from enum import Enum, unique
from typing import List, Dict

import numpy
import numpy.random

from qcompute_qnet.protocols.protocol import Protocol, SubProtocol
from qcompute_qnet.messages.message import ClassicalMessage
from qcompute_qnet.models.qpu.message import QuantumMsg

__all__ = [
    "Teleportation",
    "EntanglementSwapping",
    "BellTest",
    "CHSHGame",
    "MagicSquareGame"
]


class Teleportation(Protocol):
    r"""Class for the teleportation protocol.

    Attributes:
        role (SubProtocol): sub-protocol of a specific role
    """

    def __init__(self, name=None):
        r"""Constructor for Teleportation class.

        Args:
            name (str): name of the teleportation protocol
        """
        super().__init__(name)
        self.role = None

    class Message(ClassicalMessage):
        r"""Class for the classical control messages in teleportation protocol.
        """

        def __init__(self, src: "Node", dst: "Node", protocol: type, data: Dict):
            r"""Constructor for Message class.

            Args:
                src (Node): source of Message
                dst (Node): destination of Message
                protocol (type): protocol of Message
                data (Dict): message content
            """
            super().__init__(src, dst, protocol, data)

        @unique
        class Type(Enum):
            r"""Class for Message types.
            """

            ENT_REQUEST = "Entanglement request"
            OUTCOME_FROM_SENDER = "Measurement outcome from the sender"

    def start(self, **kwargs) -> None:
        r"""Start the teleportation protocol.

        Args:
            **kwargs: keyword arguments to start the protocol
        """
        role = kwargs['role']
        # Instantiate a sub-protocol by its role, e.g., if role == "Sender", instantiate a "Sender" sub-protocol
        self.role = getattr(Teleportation, role)(self)
        self.role.start(**kwargs)

    def receive_classical_msg(self, msg: "ClassicalMessage", **kwargs) -> None:
        r"""Receive a classical message from the node.

        Args:
            msg (ClassicalMessage): classical message used for the teleportation protocol
            **kwargs (Any): keyword arguments for receiving the classical message
        """
        self.role.receive_classical_msg(msg)

    def receive_quantum_msg(self, msg: "QuantumMsg", **kwargs) -> None:
        r"""Receive a quantum message from the node.

        Note:
            It stores the received qubit first
            and some additional receiving actions are performed based on the role of this node.

        Args:
            msg (QuantumMsg): quantum message used for the teleportation protocol
            **kwargs (Any): keyword arguments for receiving the quantum message
        """
        self.node.env.logger.debug(f"{self.node.name} received an entangled qubit from {kwargs['src'].name}")
        self.node.qreg.store_qubit(msg.data, kwargs['src'])

        self.role.receive_quantum_msg()

    class Source(SubProtocol):
        r"""Class for the role of entanglement source in teleportation protocol.
        """

        def __init__(self, super_protocol: Protocol):
            r"""Constructor for Source class.

            Args:
                super_protocol (Protocol): super protocol of the Source protocol
            """
            super().__init__(super_protocol)

        def start(self, **kwargs) -> None:
            pass

        def receive_classical_msg(self, msg: "ClassicalMessage") -> None:
            r"""Receive a classical message from the node.

            Args:
                msg (ClassicalMessage): classical message used for the teleportation protocol
            """
            if msg.data['type'] == Teleportation.Message.Type.ENT_REQUEST:
                self.node.env.logger.debug(f"{self.node.name} received an entanglement request from {msg.src.name}"
                                           f" for teleportation")
                # Entanglement generation
                self.node.qreg.h(0)
                self.node.qreg.cnot([0, 1])

                # Entanglement distribution
                self.node.send_quantum_msg(dst=msg.src, qreg_address=0)
                self.node.send_quantum_msg(dst=msg.data['peer'], qreg_address=1)

    class Sender(SubProtocol):
        r"""Class for the role of sender in teleportation protocol.

        Attributes:
            peer (QuantumNode): peer node to implement the same task
            ent_source (QuantumNode): entanglement source for entanglement distribution
            address_to_teleport (int): address of the qubit to teleport in local quantum register
        """

        def __init__(self, super_protocol: Protocol):
            r"""Constructor for Sender class.

            Args:
                super_protocol (Protocol): super protocol of the Sender protocol
            """
            super().__init__(super_protocol)
            self.peer = None
            self.ent_source = None
            self.address_to_teleport = None

        def start(self, **kwargs) -> None:
            r"""Start the teleportation protocol for the sender.

            Args:
                **kwargs: keyword arguments to start the protocol
            """
            self.peer = kwargs['peer']
            self.ent_source = kwargs['ent_source']
            self.address_to_teleport = kwargs['address_to_teleport']

            self.node.env.logger.info(f"{self.node.name} started the teleportation protocol with {self.peer.name}")
            self.request_entanglement()

        def request_entanglement(self) -> None:
            r"""Send an entanglement distribution request to the entanglement source.
            """
            ent_request_msg = Teleportation.Message(
                src=self.node, dst=self.ent_source, protocol=Teleportation,
                data={'type': Teleportation.Message.Type.ENT_REQUEST, 'peer': self.peer}
            )
            self.node.send_classical_msg(dst=self.ent_source, msg=ent_request_msg)

        def receive_quantum_msg(self) -> None:
            r"""Receive a quantum message from the node.
            """
            # Perform Bell state measurement
            address_reception = self.node.qreg.get_address(self.ent_source)  # get the address of received qubit
            self.node.qreg.bsm([self.address_to_teleport, address_reception])
            self.node.env.logger.debug(f"{self.node.name} finished local operations for teleportation")

            outcome_msg = Teleportation.Message(
                src=self.node, dst=self.peer, protocol=Teleportation,
                data={'type': Teleportation.Message.Type.OUTCOME_FROM_SENDER,
                      'outcome_from_sender': [self.node.qreg.units[self.address_to_teleport]['outcome'],
                                              self.node.qreg.units[address_reception]['outcome']]}
            )
            self.node.send_classical_msg(dst=self.peer, msg=outcome_msg)

    class Receiver(SubProtocol):
        r"""Class for the role of receiver in teleportation protocol.

        Attributes:
            peer (QuantumNode): peer node to implement the same task
            ent_source (QuantumNode): entanglement source for entanglement distribution
            outcome_from_sender (List[int]): measurement outcome received from the sender
        """

        def __init__(self, super_protocol: Protocol):
            r"""Constructor for Receiver class.

            Args:
                super_protocol (Protocol): super protocol of the Receiver protocol
            """
            super().__init__(super_protocol)
            self.peer = None
            self.ent_source = None
            self.outcome_from_sender = None

        def start(self, **kwargs) -> None:
            r"""Start the teleportation protocol for the receiver.

            Args:
                **kwargs: keyword arguments to start the protocol
            """
            self.peer = kwargs['peer']
            self.ent_source = kwargs['ent_source']

        def receive_classical_msg(self, msg: "ClassicalMessage") -> None:
            r"""Receive a classical message from the node.

            Args:
                msg (ClassicalMessage): classical message used for the teleportation protocol
            """
            if msg.data['type'] == Teleportation.Message.Type.OUTCOME_FROM_SENDER:
                self.node.env.logger.debug(f"{self.node.name} received the measurement outcome from {msg.src.name} "
                                           f"for teleportation")
                self.outcome_from_sender = msg.data['outcome_from_sender']

                address_reception = self.node.qreg.get_address(self.ent_source)
                if self.node.qreg.units[address_reception]['qubit'] is not None:
                    self.correct_state()

        def receive_quantum_msg(self) -> None:
            r"""Receive a quantum message from the node.
            """
            if self.outcome_from_sender is not None:
                self.correct_state()

        def correct_state(self) -> None:
            r"""Correct the quantum state with Pauli operators.
            """
            address_reception = self.node.qreg.get_address(self.ent_source)
            self.node.qreg.z(address_reception, condition=self.outcome_from_sender[0])
            self.node.qreg.x(address_reception, condition=self.outcome_from_sender[1])

            # After the state correction we measure the qubit state to check the teleported state
            self.node.qreg.measure(address_reception)

            self.node.env.logger.info(f"{self.node.name} finished state correction and recovered the quantum state")


class EntanglementSwapping(Protocol):
    r"""Class for the entanglement swapping protocol.

    Attributes:
        role (SubProtocol): sub-protocol of a specific role

    Note:
        Once the entanglement swapping is done, both nodes measure their qubits for verification of the protocol.
    """

    def __init__(self, name=None):
        r"""Constructor for EntanglementSwapping class.

        Args:
            name (str): name of the entanglement swapping protocol
        """
        super().__init__(name)
        self.role = None

    class Message(ClassicalMessage):
        r"""Class for the classical control messages in entanglement swapping protocol.
        """

        def __init__(self, src: "Node", dst: "Node", protocol: type, data: Dict):
            r"""Constructor for Message class.

            Args:
                src (Node): source of Message
                dst (Node): destination of Message
                protocol (type): protocol of Message
                data (Dict): message content
            """
            super().__init__(src, dst, protocol, data)

        @unique
        class Type(Enum):
            r"""Class for Message types.
            """

            SWAP_REQUEST = "Swapping request"
            ENT_REQUEST = "Entanglement request"
            OUTCOME_FROM_REPEATER = "Measurement outcome from the repeater"

    def start(self, **kwargs) -> None:
        r"""Start the entanglement swapping protocol.

        Args:
            **kwargs: keyword arguments to start the protocol
        """
        role = kwargs['role']
        self.role = getattr(EntanglementSwapping, role)(self)  # instantiate a sub-protocol by its role
        self.role.start(**kwargs)

    def receive_classical_msg(self, msg: "ClassicalMessage", **kwargs) -> None:
        r"""Receive a classical message from the node.

        Args:
            msg (ClassicalMessage): classical message used for the entanglement swapping protocol
            **kwargs (Any): keyword arguments for receiving the classical message
        """
        self.role.receive_classical_msg(msg)

    def receive_quantum_msg(self, msg: "QuantumMsg", **kwargs) -> None:
        r"""Receive a quantum message from the node.

        Note:
            It stores the received qubit first
            and some additional receiving actions are performed based on the role of this node.

        Args:
            msg (QuantumMsg): quantum message used for the entanglement swapping protocol
            **kwargs (Any): keyword arguments for receiving the quantum message
        """
        self.node.env.logger.debug(f"{self.node.name} received an entangled qubit from {kwargs['src'].name}")
        self.node.qreg.store_qubit(msg.data, kwargs['src'])

        self.role.receive_quantum_msg()

    class Source(SubProtocol):
        r"""Class for the role of entanglement source in entanglement swapping protocol.
        """

        def __init__(self, super_protocol: Protocol):
            r"""Constructor for Source class.

            Args:
                super_protocol (Protocol): super protocol of the Source protocol
            """
            super().__init__(super_protocol)

        def start(self, **kwargs) -> None:
            pass

        def receive_classical_msg(self, msg: "ClassicalMessage") -> None:
            r"""Receive a classical message from the node.

            Args:
                msg (ClassicalMessage): classical message used for the entanglement swapping protocol
            """
            if msg.data['type'] == EntanglementSwapping.Message.Type.ENT_REQUEST:
                self.node.env.logger.debug(f"{self.node.name} received an entanglement request from {msg.src.name}")
                # Entanglement generation
                self.node.qreg.h(0)
                self.node.qreg.cnot([0, 1])

                # Entanglement distribution
                self.node.send_quantum_msg(dst=msg.src, qreg_address=0)
                self.node.send_quantum_msg(dst=msg.data['peer'], qreg_address=1)

    class UpstreamNode(SubProtocol):
        r"""Class for the role of upstream node in entanglement swapping protocol.

        Attributes:
            repeater (QuantumNode): quantum repeater for entanglement swapping
            peer (QuantumNode): peer node to share entanglement
        """

        def __init__(self, super_protocol: Protocol):
            r"""Constructor for UpstreamNode class.

            Args:
                super_protocol (Protocol): super protocol of the UpstreamNode protocol
            """
            super().__init__(super_protocol)
            self.repeater = None
            self.peer = None

        def start(self, **kwargs) -> None:
            r"""Start the entanglement swapping protocol for the upstream node.

            Args:
                **kwargs: keyword arguments to start the protocol
            """
            self.peer = kwargs['peer']
            self.repeater = kwargs['repeater']
            self.node.env.logger.info(f"{self.node.name} started entanglement swapping protocol with {self.peer.name}")
            self.request_swapping()

        def request_swapping(self) -> None:
            r"""Send an entanglement swapping request to the repeater.
            """
            swap_request_msg = EntanglementSwapping.Message(
                src=self.node, dst=self.repeater, protocol=EntanglementSwapping,
                data={'type': EntanglementSwapping.Message.Type.SWAP_REQUEST, 'peer': self.peer}
            )
            self.node.send_classical_msg(dst=self.repeater, msg=swap_request_msg)

        def receive_quantum_msg(self) -> None:
            r"""Receive a quantum message from the node.
            """
            # Measure the received qubit for verification of the protocol
            self.node.qreg.measure(0)
            self.node.env.logger.debug(f"{self.node.name} measured the received qubit for verification")

    class DownstreamNode(SubProtocol):
        r"""Class for the role of downstream node in entanglement swapping protocol.

        Attributes:
            peer (QuantumNode): peer node to share entanglement
            outcome_from_repeater (List[int]): measurement outcome received from the repeater
        """

        def __init__(self, super_protocol: Protocol):
            r"""Constructor for DownstreamNode class.

            Args:
                super_protocol (Protocol): super protocol of the DownstreamNode protocol
            """
            super().__init__(super_protocol)
            self.peer = None
            self.outcome_from_repeater = None

        def start(self, **kwargs) -> None:
            r"""Start the entanglement swapping protocol for the downstream node.

            Args:
                **kwargs: keyword arguments to start the protocol
            """
            self.peer = kwargs['peer']

        def receive_classical_msg(self, msg: "ClassicalMessage") -> None:
            r"""Receive a classical message from the node.

            Args:
                msg (ClassicalMessage): classical message used for the entanglement swapping protocol
            """
            if msg.data['type'] == EntanglementSwapping.Message.Type.OUTCOME_FROM_REPEATER:
                self.node.env.logger.debug(f"{self.node.name} received the measurement outcome from {msg.src.name}")
                self.outcome_from_repeater = msg.data['outcome_from_repeater']

                if self.node.qreg.units[0]['qubit'] is not None:
                    self.correct_state()

        def receive_quantum_msg(self) -> None:
            r"""Receive a quantum message from the node.
            """
            if self.outcome_from_repeater is not None:
                self.correct_state()

        def correct_state(self) -> None:
            r"""Correct the quantum state with Pauli operators.
            """
            self.node.qreg.z(0, condition=self.outcome_from_repeater[0])
            self.node.qreg.x(0, condition=self.outcome_from_repeater[1])
            # Measure the received qubit for verification of the protocol
            self.node.qreg.measure(0)
            self.node.env.logger.debug(f"{self.node.name} measured the received qubit for verification")

    class Repeater(SubProtocol):
        r"""Class for the role of repeater in entanglement swapping protocol.

        Attributes:
            ent_sources (List[QuantumNode]): entanglement sources for distributing entanglement
            upstream_node (QuantumNode): upstream node of the entanglement swapping
            downstream_node (QuantumNode): downstream node of the entanglement swapping
        """

        def __init__(self, super_protocol: Protocol):
            r"""Constructor for Repeater class.

            Args:
                super_protocol (Protocol): super protocol of the Repeater protocol
            """
            super().__init__(super_protocol)
            self.ent_sources = None
            self.upstream_node = None
            self.downstream_node = None

        def start(self, **kwargs) -> None:
            r"""Start the entanglement swapping protocol for the repeater.

            Args:
                **kwargs: keyword arguments to start the protocol
            """
            self.ent_sources = kwargs['ent_sources']

        def request_entanglement(self, source: "Node", peer: "Node") -> None:
            r"""Send an entanglement distribution request to the entanglement source.
            """
            ent_request_msg = EntanglementSwapping.Message(
                src=self.node, dst=source, protocol=EntanglementSwapping,
                data={'type': EntanglementSwapping.Message.Type.ENT_REQUEST, 'peer': peer}
            )
            self.node.send_classical_msg(dst=source, msg=ent_request_msg)

        def receive_classical_msg(self, msg: "ClassicalMessage") -> None:
            r"""Receive a classical message from the node.

            Args:
                msg (ClassicalMessage): classical message used for the entanglement swapping protocol
            """
            if msg.data['type'] == EntanglementSwapping.Message.Type.SWAP_REQUEST:
                self.upstream_node, self.downstream_node = msg.src, msg.data['peer']

                self.request_entanglement(self.ent_sources[0], self.upstream_node)
                self.request_entanglement(self.ent_sources[1], self.downstream_node)

        def receive_quantum_msg(self) -> None:
            r"""Receive a quantum message from the node.
            """
            if self.node.qreg.units[0]['qubit'] is not None and self.node.qreg.units[1]['qubit'] is not None:
                address_0 = self.node.qreg.get_address(self.ent_sources[0])
                address_1 = self.node.qreg.get_address(self.ent_sources[1])
                self.node.qreg.bsm([address_0, address_1])

                outcome_msg = EntanglementSwapping.Message(
                    src=self.node, dst=self.downstream_node, protocol=EntanglementSwapping,
                    data={'type': EntanglementSwapping.Message.Type.OUTCOME_FROM_REPEATER,
                          'outcome_from_repeater': [self.node.qreg.units[address_0]['outcome'],
                                                    self.node.qreg.units[address_1]['outcome']]
                          }
                )
                self.node.send_classical_msg(dst=self.downstream_node, msg=outcome_msg)


class BellTest(Protocol):
    r"""Class for the BellTest protocol.

    Attributes:
        role (SubProtocol): sub-protocol of a specific role
    """

    def __init__(self, name=None):
        r"""Constructor for BellTest class.

        Args:
            name (str): name of the BellTest protocol
        """
        super().__init__(name)
        self.role = None

    def start(self, **kwargs) -> None:
        r"""Start the BellTest protocol.

        Args:
            **kwargs: keyword arguments to start the protocol
        """
        role = kwargs['role']
        self.role = getattr(BellTest, role)(self)  # instantiate a sub-protocol by its role
        self.role.start(**kwargs)

    def receive_classical_msg(self, msg: "ClassicalMessage", **kwargs) -> None:
        r"""Receive a classical message from the node.

        Args:
            msg (ClassicalMessage): classical message used for the BellTest protocol
            **kwargs (Any): keyword arguments for receiving the classical message
        """
        self.role.receive_classical_msg(msg)

    def receive_quantum_msg(self, msg: "QuantumMsg", **kwargs) -> None:
        r"""Receive a quantum message from the node.

        Note:
            It stores the received qubit first
            and some additional receiving actions are performed based on the role of this node.

        Args:
            msg (QuantumMsg): quantum message used for the BellTest protocol
            **kwargs (Any): keyword arguments for receiving the quantum message
        """
        self.node.env.logger.debug(f"{self.node.name} received an entangled qubit from {kwargs['src'].name}")
        self.node.qreg.store_qubit(msg.data, kwargs['src'])
        self.node.qreg.circuit_index = msg.index  # synchronize the quantum circuit

        self.role.receive_quantum_msg()

    @staticmethod
    def estimate_statistics(results: List[Dict]) -> None:
        r"""Calculate the sum of the expectations.

        Args:
            results (List[Dict]): sample results of the circuits
        """
        print("\n" + "-" * 40)
        for i, result in enumerate(results):
            cir_name = result['circuit_name']
            shots = result['shots']
            counts = result['counts']

            if "QS" in cir_name or "SQ" in cir_name:
                avg_qs = (counts['00'] - counts['01'] - counts['10'] + counts['11']) / shots
                print(f"Average value of QS: {avg_qs:.4f}")
            elif "QT" in cir_name or "TQ" in cir_name:
                avg_qt = (counts['00'] - counts['01'] - counts['10'] + counts['11']) / shots
                print(f"Average value of QT: {avg_qt:.4f}")
            elif "RS" in cir_name or "SR" in cir_name:
                avg_rs = (counts['00'] - counts['01'] - counts['10'] + counts['11']) / shots
                print(f"Average value of RS: {avg_rs:.4f}")
            elif "RT" in cir_name or "TR" in cir_name:
                avg_rt = (counts['00'] - counts['01'] - counts['10'] + counts['11']) / shots
                print(f"Average value of RT: {avg_rt:.4f}")

        avg_sum = - avg_qs - avg_qt - avg_rs + avg_rt
        print(f"\n- <QS> - <QT> - <RS> + <RT> = {avg_sum:.4f}\n" + '-' * 40)

    class Sender(SubProtocol):
        r"""Class for the role of sender in BellTest protocol.

        Attributes:
            receivers (List[QuantumNode]): receivers of the qubits
            rounds (int): rounds for testing the Bell inequality
        """

        def __init__(self, super_protocol: Protocol):
            r"""Constructor for Sender class.

            Args:
                super_protocol (Protocol): super protocol of the Sender protocol
            """
            super().__init__(super_protocol)
            self.receivers = None
            self.rounds = None

        def start(self, **kwargs) -> None:
            r"""Start the BellTest protocol for the sender.

            Args:
                **kwargs: keyword arguments to start the protocol
            """
            self.receivers = kwargs['receivers']
            self.rounds = kwargs['rounds']

            self.node.env.logger.info(
                f"{self.node.name} started BellTest protocol with {self.receivers[0].name} and {self.receivers[1].name}"
            )
            self.distribute_entanglement()

        def distribute_entanglement(self) -> None:
            r"""Prepare and distribute pairs of entangled qubits.
            """
            for i in range(self.rounds):
                self.node.qreg.create_circuit('CHSH_')

                self.node.qreg.h(0)
                self.node.qreg.cnot([0, 1])
                self.node.qreg.z(0)
                self.node.qreg.x(0)

                self.node.send_quantum_msg(dst=self.receivers[0], qreg_address=0)
                self.node.send_quantum_msg(dst=self.receivers[1], qreg_address=1)

    class ReceiverA(SubProtocol):
        r"""Class for the role of receiver A in BellTest protocol.
        """

        def __init__(self, super_protocol: Protocol):
            r"""Constructor for ReceiverA class.

            Args:
                super_protocol (Protocol): super protocol of the ReceiverA protocol
            """
            super().__init__(super_protocol)

        def start(self, **kwargs) -> None:
            pass

        def receive_quantum_msg(self) -> None:
            r"""Receive a quantum message from the node.
            """
            measurement_choice = numpy.random.choice([0, 1])

            if measurement_choice == 0:
                self.node.qreg.measure(0, basis="z")
                self.node.qreg.circuit.name += "Q"
            else:
                self.node.qreg.measure(0, basis="x")
                self.node.qreg.circuit.name += "R"

    class ReceiverB(SubProtocol):
        r"""Class for the role of receiver B in BellTest protocol.
        """

        def __init__(self, super_protocol: Protocol):
            r"""Constructor for ReceiverB class.

            Args:
                super_protocol (Protocol): super protocol of the ReceiverB protocol
            """
            super().__init__(super_protocol)

        def start(self, **kwargs) -> None:
            pass

        def receive_quantum_msg(self) -> None:
            r"""Receive a quantum message from the node.
            """
            measurement_choice = numpy.random.choice([0, 1])

            if measurement_choice == 0:
                self.node.qreg.ry(0, - numpy.pi / 4)  # (Z + X) / \sqrt{2}
                self.node.qreg.measure(0, basis="z")
                self.node.qreg.circuit.name += "S"
            else:
                self.node.qreg.ry(0, numpy.pi / 4)  # (Z - X) / \sqrt{2}
                self.node.qreg.measure(0, basis="z")
                self.node.qreg.circuit.name += "T"


class CHSHGame(Protocol):
    r"""Class for the CHSH game protocol.

    Attributes:
        role (SubProtocol): sub-protocol of a specific role
    """

    def __init__(self, name=None):
        r"""Constructor for CHSHGame class.

        Args:
            name (str): name of the CHSHGame protocol
        """
        super().__init__(name)
        self.role = None

    class Message(ClassicalMessage):
        r"""Class for the classical control messages in CHSHGame protocol.
        """

        def __init__(self, src: "Node", dst: "Node", protocol: type, data: Dict):
            r"""Constructor for Message class.

            Args:
                src (Node): source of Message
                dst (Node): destination of Message
                protocol (type): protocol of Message
                data (Dict): message content
            """
            super().__init__(src, dst, protocol, data)

        @unique
        class Type(Enum):
            r"""Class for Message types.
            """

            ENT_REQUEST = "Entanglement request"
            READY = "Ready"
            QUESTION = "Question from the referee"
            ANSWER = "Answer from the player"

    def start(self, **kwargs) -> None:
        r"""Start the CHSHGame protocol.

        Args:
            **kwargs: keyword arguments to start the protocol
        """
        role = kwargs['role']
        self.role = getattr(CHSHGame, role)(self)  # instantiate a sub-protocol by its role
        self.role.start(**kwargs)

    def receive_classical_msg(self, msg: "ClassicalMessage", **kwargs) -> None:
        r"""Receive a classical message from the node.

        Args:
            msg (ClassicalMessage): classical message used for the CHSHGame protocol
            **kwargs (Any): keyword arguments for receiving the classical message
        """
        self.role.receive_classical_msg(msg)

    def receive_quantum_msg(self, msg: "QuantumMsg", **kwargs) -> None:
        r"""Receive a quantum message from the node.

        Note:
            It stores the received qubit first
            and some additional receiving actions are performed based on the role of this node.

        Args:
            msg (QuantumMsg): quantum message used for the CHSHGame protocol
            **kwargs (Any): keyword arguments for receiving the quantum message
        """
        self.node.env.logger.debug(f"{self.node.name} received an entangled qubit from {kwargs['src'].name}")
        self.node.qreg.store_qubit(msg.data, kwargs['src'])
        self.node.qreg.circuit_index = msg.index  # synchronize the quantum circuit

        self.role.receive_quantum_msg()

    def estimate_statistics(self, results: List[Dict]) -> None:
        r"""Calculate the winning probability of the CHSH game.

        Args:
            results (List[Dict]): sample results of the circuits
        """
        assert type(self.role).__name__ == "Referee", \
            f"The role of {type(self.role).__name__} has no right to calculate the winning probability of the game!"

        self.role.estimate_statistics(results)

    class Source(SubProtocol):
        r"""Class for the role of entanglement source in the CHSHGame protocol.
        """

        def __init__(self, super_protocol: Protocol):
            r"""Constructor for Source class.

            Args:
                super_protocol (Protocol): super protocol of the Source protocol
            """
            super().__init__(super_protocol)

        def start(self, **kwargs) -> None:
            pass

        def receive_classical_msg(self, msg: "ClassicalMessage") -> None:
            r"""Receive a classical message from the node.

            Args:
                msg (ClassicalMessage): classical message used for the CHSHGame protocol
            """
            if msg.data['type'] == CHSHGame.Message.Type.ENT_REQUEST:
                self.node.env.logger.debug(f"{self.node.name} received an entanglement request from {msg.src.name}")

                self.node.qreg.create_circuit(f"CHSHGame_")
                # Entanglement generation
                self.node.qreg.h(0)
                self.node.qreg.cnot([0, 1])

                # Entanglement distribution
                self.node.send_quantum_msg(dst=msg.src, qreg_address=0)
                self.node.send_quantum_msg(dst=msg.data['peer'], qreg_address=1)

    class Player1(SubProtocol):
        r"""Class for the role of player 1 in the CHSHGame protocol.

        Attributes:
            peer (Node): another player of the CHSH game
            ent_source (Node): entanglement source for distributing entanglement
            referee (Node): referee of the CHSH game
            rounds (int): number of the game rounds
            current_round (int): current round of the game
        """

        def __init__(self, super_protocol: Protocol):
            r"""Constructor for Player1 class.

            Args:
                super_protocol (Protocol): super protocol of the Player1 protocol
            """
            super().__init__(super_protocol)
            self.peer = None
            self.ent_source = None
            self.referee = None
            self.rounds = None
            self.current_round = 0

        def start(self, **kwargs) -> None:
            r"""Start the CHSHGame protocol for the player 1.

            Args:
                **kwargs: keyword arguments to start the protocol
            """
            self.peer = kwargs['peer']
            self.ent_source = kwargs['ent_source']
            self.referee = kwargs['referee']
            self.rounds = kwargs['rounds']

            self.node.env.logger.info("Loading CHSH game...")
            self.node.env.logger.info(
                f"Referee: {self.referee.name}, Player1: {self.node.name}, Player2: {self.peer.name}"
            )
            self.prepare_for_game()

        def prepare_for_game(self) -> None:
            r"""Player 1 prepares for the CHSH game.
            """
            self.current_round += 1
            self.request_entanglement()

        def request_entanglement(self) -> None:
            r"""Send an entanglement distribution request to the entanglement source.
            """
            ent_request_msg = CHSHGame.Message(
                src=self.node, dst=self.ent_source, protocol=CHSHGame,
                data={'type': CHSHGame.Message.Type.ENT_REQUEST, 'peer': self.peer}
            )
            self.node.send_classical_msg(dst=self.ent_source, msg=ent_request_msg)

        def receive_classical_msg(self, msg: "ClassicalMessage") -> None:
            r"""Receive a classical message from the node.

            Args:
                msg (ClassicalMessage): classical message used for the CHSHGame protocol
            """
            if msg.data['type'] == CHSHGame.Message.Type.QUESTION:
                self.node.env.logger.debug(f"{self.node.name} received the question from {self.referee.name}")

                x = msg.data['question']

                if x == 0:
                    self.node.qreg.measure(0, basis="z")
                    self.node.qreg.circuit.name += "x0"

                elif x == 1:
                    self.node.qreg.measure(0, basis="x")
                    self.node.qreg.circuit.name += "x1"

                answer_msg = CHSHGame.Message(
                    src=self.node, dst=self.referee, protocol=CHSHGame,
                    data={'type': CHSHGame.Message.Type.ANSWER,
                          'answer': self.node.qreg.units[0]['outcome']}
                )
                self.node.send_classical_msg(dst=self.referee, msg=answer_msg)

                if self.current_round < self.rounds:
                    self.prepare_for_game()

        def receive_quantum_msg(self) -> None:
            r"""Receive a quantum message from the node.
            """
            self.play_game()

        def play_game(self) -> None:
            r"""Send a READY message to the referee.
            """
            ready_msg = CHSHGame.Message(
                src=self.node, dst=self.referee, protocol=CHSHGame,
                data={'type': CHSHGame.Message.Type.READY}
            )
            self.node.send_classical_msg(dst=self.referee, msg=ready_msg)

    class Player2(SubProtocol):
        r"""Class for the role of player 2 in the CHSHGame protocol.

        Attributes:
            peer (Node): another player of the CHSH game
            ent_source (Node): entanglement source for distributing entanglement
            referee (Node): referee of the CHSH game
        """

        def __init__(self, super_protocol: Protocol):
            r"""Constructor for Player2 class.

            Args:
                super_protocol (Protocol): super protocol of the Player2 protocol
            """
            super().__init__(super_protocol)
            self.peer = None
            self.ent_source = None
            self.referee = None

        def start(self, **kwargs) -> None:
            r"""Start the CHSHGame protocol for the player 2.

            Args:
                **kwargs: keyword arguments to start the protocol
            """
            self.peer = kwargs['peer']
            self.ent_source = kwargs['ent_source']
            self.referee = kwargs['referee']

        def receive_classical_msg(self, msg: "ClassicalMessage") -> None:
            r"""Receive a classical message from the node.

            Args:
                msg (ClassicalMessage): classical message used for the CHSHGame protocol
            """
            if msg.data['type'] == CHSHGame.Message.Type.QUESTION:
                self.node.env.logger.debug(f"{self.node.name} received the question from {self.referee.name}")

                y = msg.data['question']

                if y == 0:
                    self.node.qreg.ry(0, - numpy.pi / 4)  # (Z + X) / \sqrt{2}
                    self.node.qreg.measure(0, basis="z")
                    self.node.qreg.circuit.name += "y0"

                elif y == 1:
                    self.node.qreg.ry(0, numpy.pi / 4)  # (Z - X) / \sqrt{2}
                    self.node.qreg.measure(0, basis="z")
                    self.node.qreg.circuit.name += "y1"

                answer_msg = CHSHGame.Message(
                    src=self.node, dst=self.referee, protocol=CHSHGame,
                    data={'type': CHSHGame.Message.Type.ANSWER,
                          'answer': self.node.qreg.units[0]['outcome']}
                )
                self.node.send_classical_msg(dst=self.referee, msg=answer_msg)

        def receive_quantum_msg(self) -> None:
            r"""Receive a quantum message from the node.
            """
            self.play_game()

        def play_game(self) -> None:
            r"""Send a READY message to the referee.
            """
            ready_msg = CHSHGame.Message(
                src=self.node, dst=self.referee, protocol=CHSHGame,
                data={'type': CHSHGame.Message.Type.READY}
            )
            self.node.send_classical_msg(dst=self.referee, msg=ready_msg)

    class Referee(SubProtocol):
        r"""Class for the role of referee in the CHSHGame protocol.

        Attributes:
            players (List[QuantumNode]): players of the CHSH game
            players_ready (List[bool]): list for checking if both players are ready for the game
            questions (List[list]): record questions for the two players generated in each game round
            answers_p1 (list): record the answer from player 1 in each game round
            answers_p2 (list): record the answer from player 2 in each game round
        """

        def __init__(self, super_protocol: Protocol):
            r"""Constructor for Referee class.

            Args:
                super_protocol (Protocol): super protocol of the Referee protocol
            """
            super().__init__(super_protocol)
            self.players = None
            self.players_ready = [False, False]
            self.questions = []
            self.answers_p1 = []
            self.answers_p2 = []

        def start(self, **kwargs) -> None:
            r"""Start the CHSHGame protocol for the referee.

            Args:
                **kwargs: keyword arguments to start the protocol
            """
            self.players = kwargs['players']

        def receive_classical_msg(self, msg: "ClassicalMessage") -> None:
            r"""Receive a classical message from the node.

            Args:
                msg (ClassicalMessage): classical message used for the CHSHGame protocol
            """
            if msg.data['type'] == CHSHGame.Message.Type.READY:
                self.node.env.logger.debug(f"{msg.src.name} was ready for the CHSH game")
                self.players_ready[self.players.index(msg.src)] = True

                if all(self.players_ready):
                    self.node.env.logger.debug(f"All players were ready. CHSH game started")
                    self.send_questions()

                    self.players_ready = [False, False]  # reset the player's status for the next round

            elif msg.data['type'] == CHSHGame.Message.Type.ANSWER:
                self.node.env.logger.debug(f"{self.node.name} received the answer from {msg.src.name}")

                if msg.src == self.players[0]:
                    self.answers_p1.append(msg.data['answer'])
                elif msg.src == self.players[1]:
                    self.answers_p2.append(msg.data['answer'])

        def send_questions(self) -> None:
            r"""Randomly choose two questions and send them to the two players.
            """
            questions = numpy.random.choice([0, 1], size=2)  # randomly generate two bits as questions
            self.questions.append(questions)

            for i, player in enumerate(self.players):
                question_msg = CHSHGame.Message(
                    src=self.node, dst=player, protocol=CHSHGame,
                    data={'type': CHSHGame.Message.Type.QUESTION,
                          'question': questions[i]}
                )
                self.node.send_classical_msg(dst=player, msg=question_msg)

        def estimate_statistics(self, results: List[Dict]) -> None:
            r"""Calculate the winning probability of the CHSH game.

            Args:
                results (List[Dict]): sample results of the circuits
            """
            num_wins = 0

            for result in results:
                cir_name = result['circuit_name']
                counts = result['counts']

                if "x1" in cir_name and "y1" in cir_name:  # both questions are 1
                    for count in counts:
                        answer_p1, answer_p2 = self.answers_p1[0], self.answers_p2[0]
                        if int(count[answer_p1]) ^ int(count[answer_p2]) == 1:
                            num_wins += counts[count]
                else:
                    for count in counts:
                        answer_p1, answer_p2 = self.answers_p1[0], self.answers_p2[0]
                        if int(count[answer_p1]) ^ int(count[answer_p2]) == 0:
                            num_wins += counts[count]

            winning_prob = num_wins / sum(result['shots'] for result in results)
            print(f"\n{'-' * 55}\nThe winning probability of the CHSH game is {winning_prob:.4f}.\n{'-' * 55}")


class MagicSquareGame(Protocol):
    r"""Class for the MagicSquareGame protocol.

    Attributes:
        role (SubProtocol): sub-protocol of a specific role
    """

    def __init__(self, name=None):
        r"""Constructor for MagicSquareGame class.

        Args:
            name (str): name of the MagicSquareGame protocol
        """
        super().__init__(name)
        self.role = None

    class Message(ClassicalMessage):
        r"""Class for the classical control messages in MagicSquareGame protocol.
        """

        def __init__(self, src: "Node", dst: "Node", protocol: type, data: Dict):
            r"""Constructor for Message class.

            Args:
                src (Node): source of Message
                dst (Node): destination of Message
                protocol (type): protocol of Message
                data (Dict): message content
            """
            super().__init__(src, dst, protocol, data)

        @unique
        class Type(Enum):
            r"""Class for Message types.
            """

            ENT_REQUEST = "Entanglement request"
            READY = "Ready"
            QUESTION = "Question from the referee"
            ANSWER = "Answer from the player"

    def start(self, **kwargs) -> None:
        r"""Start the MagicSquareGame protocol.

        Args:
            **kwargs: keyword arguments to start the protocol
        """
        role = kwargs['role']
        self.role = getattr(MagicSquareGame, role)(self)  # instantiate a sub-protocol by its role
        self.role.start(**kwargs)

    def receive_classical_msg(self, msg: "ClassicalMessage", **kwargs) -> None:
        r"""Receive a classical message from the node.

        Args:
            msg (ClassicalMessage): classical message used for the MagicSquareGame protocol
            **kwargs (Any): keyword arguments for receiving the classical message
        """
        self.role.receive_classical_msg(msg)

    def receive_quantum_msg(self, msg: "QuantumMsg", **kwargs) -> None:
        r"""Receive a quantum message from the node.

        Note:
            It stores the received qubit first
            and some additional receiving actions are performed based on the role of this node.

        Args:
            msg (QuantumMsg): quantum message used for the MagicSquareGame protocol
            **kwargs (Any): keyword arguments for receiving the quantum message
        """
        self.node.env.logger.debug(f"{self.node.name} received an entangled qubit from {kwargs['src'].name}")
        self.node.qreg.store_qubit(msg.data, kwargs['src'])
        self.node.qreg.circuit_index = msg.index  # synchronize the quantum circuit

        self.role.receive_quantum_msg()

    def estimate_statistics(self, results: List[Dict]) -> None:
        r"""Calculate the winning probability of the magic square game.

        Args:
            results (List[Dict]): sample results of the circuits
        """
        assert type(self.role).__name__ == "Referee",\
            f"The role of {type(self.role).__name__} has no right to calculate the winning probability of the game!"

        self.role.estimate_statistics(results)

    class Source(SubProtocol):
        r"""Class for the role of entanglement source in the MagicSquareGame protocol.
        """

        def __init__(self, super_protocol: Protocol):
            r"""Constructor for Source class.

            Args:
                super_protocol (Protocol): super protocol of the Source protocol
            """
            super().__init__(super_protocol)

        def start(self, **kwargs) -> None:
            pass

        def receive_classical_msg(self, msg: "ClassicalMessage") -> None:
            r"""Receive a classical message from the node.

            Args:
                msg (ClassicalMessage): classical message used for the MagicSquareGame protocol
            """
            if msg.data['type'] == MagicSquareGame.Message.Type.ENT_REQUEST:
                self.node.env.logger.debug(f"{self.node.name} received an entanglement request from {msg.src.name}")

                self.node.qreg.create_circuit(f"MagicSquareGame_")
                # Entanglement generation
                self.node.qreg.h(0)
                self.node.qreg.cnot([0, 1])
                self.node.qreg.h(2)
                self.node.qreg.cnot([2, 3])

                # Entanglement distribution
                self.node.send_quantum_msg(dst=msg.src, qreg_address=0, priority=0)
                self.node.send_quantum_msg(dst=msg.src, qreg_address=2, priority=1)
                self.node.send_quantum_msg(dst=msg.data['peer'], qreg_address=1, priority=0)
                self.node.send_quantum_msg(dst=msg.data['peer'], qreg_address=3, priority=1)

    class Player1(SubProtocol):
        r"""Class for the role of player 1 in the MagicSquareGame protocol.

        Attributes:
            peer (Node): another player of the magic square game
            ent_source (Node): entanglement source for distributing entanglement
            referee (Node): referee of the magic square game
            rounds (int): number of the game rounds
            current_round (int): current round of the game
        """

        def __init__(self, super_protocol: Protocol):
            r"""Constructor for Player1 class.

            Args:
                super_protocol (Protocol): super protocol of the Player1 protocol
            """
            super().__init__(super_protocol)
            self.peer = None
            self.ent_source = None
            self.referee = None
            self.rounds = None
            self.current_round = 0

        def start(self, **kwargs) -> None:
            r"""Start the MagicSquareGame protocol for the player 1.

            Args:
                **kwargs: keyword arguments to start the protocol
            """
            self.peer = kwargs['peer']
            self.ent_source = kwargs['ent_source']
            self.referee = kwargs['referee']
            self.rounds = kwargs['rounds']

            self.node.env.logger.info("Loading magic square game...")
            self.node.env.logger.info(
                f"Referee: {self.referee.name}, Player1: {self.node.name}, Player2: {self.peer.name}"
            )
            self.prepare_for_game()

        def prepare_for_game(self) -> None:
            r"""Player 1 prepares for the magic square game.
            """
            self.current_round += 1
            self.request_entanglement()

        def request_entanglement(self) -> None:
            r"""Send an entanglement distribution request to the entanglement source.
            """
            ent_request_msg = MagicSquareGame.Message(
                src=self.node, dst=self.ent_source, protocol=MagicSquareGame,
                data={'type': MagicSquareGame.Message.Type.ENT_REQUEST, 'peer': self.peer}
            )
            self.node.send_classical_msg(dst=self.ent_source, msg=ent_request_msg)

        def receive_classical_msg(self, msg: "ClassicalMessage") -> None:
            r"""Receive a classical message from the node.

            Args:
                msg (ClassicalMessage): classical message used for the MagicSquareGame protocol
            """
            if msg.data['type'] == MagicSquareGame.Message.Type.QUESTION:
                self.node.env.logger.debug(f"{self.node.name} received the question from {self.referee.name}")

                row = msg.data['question']

                if row == 0:
                    self.node.qreg.measure(0, basis="z")
                    self.node.qreg.measure(1, basis="z")
                    self.node.qreg.circuit.name += "r0"

                elif row == 1:
                    self.node.qreg.measure(0, basis="x")
                    self.node.qreg.measure(1, basis="x")
                    self.node.qreg.circuit.name += "r1"

                elif row == 2:
                    self.node.qreg.z(0)
                    self.node.qreg.z(1)
                    self.node.qreg.cz([0, 1])
                    self.node.qreg.h(0)
                    self.node.qreg.h(1)
                    self.node.qreg.measure(0, basis="z")
                    self.node.qreg.measure(1, basis="z")
                    self.node.qreg.circuit.name += "r2"

                answer_msg = MagicSquareGame.Message(
                    src=self.node, dst=self.referee, protocol=MagicSquareGame,
                    data={'type': MagicSquareGame.Message.Type.ANSWER,
                          'answer': [self.node.qreg.units[0]['outcome'],
                                     self.node.qreg.units[1]['outcome']]
                          }
                )
                self.node.send_classical_msg(dst=self.referee, msg=answer_msg)

                if self.current_round < self.rounds:
                    self.prepare_for_game()

        def receive_quantum_msg(self) -> None:
            r"""Receive a quantum message from the node.
            """
            if all(unit['qubit'] is not None for unit in self.node.qreg.units):
                self.play_game()

        def play_game(self) -> None:
            r"""Send a READY message to the referee.
            """
            ready_msg = MagicSquareGame.Message(
                src=self.node, dst=self.referee, protocol=MagicSquareGame,
                data={'type': MagicSquareGame.Message.Type.READY}
            )
            self.node.send_classical_msg(dst=self.referee, msg=ready_msg)

    class Player2(SubProtocol):
        r"""Class for the role of player 2 in the MagicSquareGame protocol.

        Attributes:
            peer (Node): another player of the magic square game
            ent_source (Node): entanglement source for distributing entanglement
            referee (Node): referee of the magic square game
        """

        def __init__(self, super_protocol: Protocol):
            r"""Constructor for Player2 class.

            Args:
                super_protocol (Protocol): super protocol of the Player2 protocol
            """
            super().__init__(super_protocol)
            self.peer = None
            self.ent_source = None
            self.referee = None

        def start(self, **kwargs) -> None:
            r"""Start the MagicSquareGame protocol for the player 2.

            Args:
                **kwargs: keyword arguments to start the protocol
            """
            self.peer = kwargs['peer']
            self.ent_source = kwargs['ent_source']
            self.referee = kwargs['referee']

        def receive_classical_msg(self, msg: "ClassicalMessage") -> None:
            r"""Receive a classical message from the node.

            Args:
                msg (ClassicalMessage): classical message used for the MagicSquareGame protocol
            """
            if msg.data['type'] == MagicSquareGame.Message.Type.QUESTION:
                self.node.env.logger.debug(f"{self.node.name} received the question from {self.referee.name}")

                column = msg.data['question']

                if column == 0:
                    self.node.qreg.measure(0, basis="x")
                    self.node.qreg.measure(1, basis="z")
                    self.node.qreg.circuit.name += "c0"

                elif column == 1:
                    self.node.qreg.measure(0, basis="z")
                    self.node.qreg.measure(1, basis="x")
                    self.node.qreg.circuit.name += "c1"

                elif column == 2:
                    self.node.qreg.bsm([0, 1])
                    self.node.qreg.circuit.name += "c2"

                answer_msg = MagicSquareGame.Message(
                    src=self.node, dst=self.referee, protocol=MagicSquareGame,
                    data={'type': MagicSquareGame.Message.Type.ANSWER,
                          'answer': [self.node.qreg.units[0]['outcome'],
                                     self.node.qreg.units[1]['outcome']]
                          }
                )
                self.node.send_classical_msg(dst=self.referee, msg=answer_msg)

        def receive_quantum_msg(self) -> None:
            r"""Receive a quantum message from the node.
            """
            if all(unit['qubit'] is not None for unit in self.node.qreg.units):
                self.play_game()

        def play_game(self) -> None:
            r"""Send a READY message to the referee.
            """
            ready_msg = MagicSquareGame.Message(
                src=self.node, dst=self.referee, protocol=MagicSquareGame,
                data={'type': MagicSquareGame.Message.Type.READY}
            )
            self.node.send_classical_msg(dst=self.referee, msg=ready_msg)

    class Referee(SubProtocol):
        r"""Class for the role of referee in the MagicSquareGame protocol.

        Attributes:
            players (List[QuantumNode]): players of the magic square game
            players_ready (List[bool]): list for checking if both players are ready for the game
            questions (List[list]): record questions for the two players generated in each game round
            answers_p1 (list): record the answer from player 1 in each game round
            answers_p2 (list): record the answer from player 2 in each game round
        """

        def __init__(self, super_protocol: Protocol):
            r"""Constructor for Referee class.

            Args:
                super_protocol (Protocol): super protocol of the Referee protocol
            """
            super().__init__(super_protocol)
            self.players = None
            self.players_ready = [False, False]
            self.questions = []
            self.answers_p1 = []
            self.answers_p2 = []

        def start(self, **kwargs) -> None:
            r"""Start the MagicSquareGame protocol for the referee.

            Args:
                **kwargs: keyword arguments to start the protocol
            """
            self.players = kwargs['players']

        def receive_classical_msg(self, msg: "ClassicalMessage") -> None:
            r"""Receive a classical message from the node.

            Args:
                msg (ClassicalMessage): classical message used for the MagicSquareGame protocol
            """
            if msg.data['type'] == MagicSquareGame.Message.Type.READY:
                self.node.env.logger.debug(f"{msg.src.name} was ready for the magic square game")
                self.players_ready[self.players.index(msg.src)] = True

                if all(self.players_ready):
                    self.node.env.logger.debug(f"All players were ready. Magic square game started")
                    self.send_questions()

                    self.players_ready = [False, False]  # reset the players' status for the next round

            elif msg.data['type'] == MagicSquareGame.Message.Type.ANSWER:
                self.node.env.logger.debug(f"{self.node.name} received the answer from {msg.src.name}")

                if msg.src == self.players[0]:
                    self.answers_p1.append(msg.data['answer'])
                elif msg.src == self.players[1]:
                    self.answers_p2.append(msg.data['answer'])

        def send_questions(self) -> None:
            r"""Randomly choose the row and the column to fill and send the indices to the two players respectively.
            """
            questions = numpy.random.choice([0, 1, 2], size=2)
            self.questions.append(questions)

            for i, player in enumerate(self.players):
                question_msg = MagicSquareGame.Message(
                    src=self.node, dst=self.players[i], protocol=MagicSquareGame,
                    data={'type': MagicSquareGame.Message.Type.QUESTION,
                          'question': questions[i]}
                )
                self.node.send_classical_msg(dst=self.players[i], msg=question_msg)

        def estimate_statistics(self, results: List[Dict]) -> None:
            r"""Calculate the winning probability of the magic square game.

            Args:
                results (List[Dict]): sample results of the circuits
            """
            def p1_answer(row: int, outcome: List[int]) -> list:
                r"""Answer of the player 1.

                Args:
                    row (int): row index
                    outcome (List[int]): measurement outcome of the player 1

                Returns:
                    list: numbers to fill the row
                """
                if row == 0:
                    if outcome == [0, 0]:
                        return [1, 1, 1]
                    elif outcome == [0, 1]:
                        return [-1, 1, -1]
                    elif outcome == [1, 0]:
                        return [1, -1, -1]
                    elif outcome == [1, 1]:
                        return [-1, -1, 1]

                elif row == 1 or row == 2:
                    if outcome == [0, 0]:
                        return [1, 1, 1]
                    elif outcome == [0, 1]:
                        return [1, -1, -1]
                    elif outcome == [1, 0]:
                        return [-1, 1, -1]
                    elif outcome == [1, 1]:
                        return [-1, -1, 1]

            def p2_answer(column: int, outcome: List[int]) -> list:
                r"""Answer of the player 2.

                Args:
                    column (int): column index
                    outcome (List[int]): measurement outcome of the player 2

                Returns:
                    list: numbers to fill the column
                """
                if column == 0 or column == 2:
                    if outcome == [0, 0]:
                        return [1, 1, -1]
                    elif outcome == [0, 1]:
                        return [-1, 1, 1]
                    elif outcome == [1, 0]:
                        return [1, -1, 1]
                    elif outcome == [1, 1]:
                        return [-1, -1, -1]

                if column == 1:
                    if outcome == [0, 0]:
                        return [1, 1, -1]
                    elif outcome == [0, 1]:
                        return [1, -1, 1]
                    elif outcome == [1, 0]:
                        return [-1, 1, 1]
                    elif outcome == [1, 1]:
                        return [-1, -1, -1]

            def is_winning(row: int, column: int, answer_p1: List[int], answer_p2: List[int]) -> bool:
                r"""Determine if the players win the game.

                args:
                    row (int): row index
                    column (int): column index
                    answer_p1 (list[int]): answer of the player 1
                    answer_p2 (List[int]): answer of the player 2

                Returns:
                    bool: whether the players win
                """
                if numpy.prod(answer_p1) == 1 and numpy.prod(answer_p2) == -1 and answer_p1[column] == answer_p2[row]:
                    return True
                else:
                    return False

            num_wins = 0

            for i, result in enumerate(results):
                cir_name = result['circuit_name']
                counts = result['counts']

                if "r0" in cir_name:
                    row = 0
                elif "r1" in cir_name:
                    row = 1
                elif "r2" in cir_name:
                    row = 2

                if "c0" in cir_name:
                    column = 0
                elif "c1" in cir_name:
                    column = 1
                elif "c2" in cir_name:
                    column = 2

                for count in counts:
                    answer_p1, answer_p2 = self.answers_p1[0], self.answers_p2[0]
                    outcome_p1 = [int(count[answer_p1[0]]), int(count[answer_p1[1]])]
                    outcome_p2 = [int(count[answer_p2[0]]), int(count[answer_p2[1]])]

                    if is_winning(row, column, p1_answer(row, outcome_p1), p2_answer(column, outcome_p2)):
                        num_wins += counts[count]

            winning_prob = num_wins / sum(result['shots'] for result in results)
            print(f"\n{'-' * 60}\nThe winning probability of the magic square game is {winning_prob:.4f}.\n{'-' * 60}")



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for photon sources.
"""

from typing import List
from numpy import random
from qcompute_qnet.core.des import Entity, EventHandler
from qcompute_qnet.devices.photon import Photon
from qcompute_qnet.quantum.state import MixedState

__all__ = [
    "PhotonSource"
]


class PhotonSource(Entity):
    r"""Class for the simulation of a photon source.

    Attributes:
        frequency (float): frequency at which light pulses are emitted in hertz
        wavelength (float): wavelength of photons in nanometers
        bandwidth (float): bandwidth of photon wavelength in nanometers
        mean_photon_num (float): average number of photons per signal pulse
    """

    def __init__(self, name: str, env=None, frequency=80e6, wavelength=1550, bandwidth=0, mean_photon_num=0.5):
        r"""Constructor for the PhotonSource class.

        Args:
            name (str): name of the photon source
            env (DESEnv): discrete-event simulation environment
            frequency (float): frequency at which light pulses are emitted in hertz
            wavelength (float): wavelength of photons in nanometers
            bandwidth (float): bandwidth of photon wavelength in nanometers
            mean_photon_num (float): average number of photons per signal pulse
        """
        super().__init__(name, env)
        self.frequency = frequency
        self.wavelength = wavelength
        self.bandwidth = bandwidth
        self.mean_photon_num = mean_photon_num

    def init(self) -> None:
        r"""Photon source initialization.
        """
        assert self.owner != self, f"The photon source {self.name} has no owner!"

    def set(self, **kwargs) -> None:
        r"""Set given parameters for the photon source.

        Args:
            **kwargs: keyword arguments to set
        """
        for attr in kwargs:
            if attr == "frequency":
                assert isinstance(kwargs[attr], float), "'frequency' should be a float value."
                self.__setattr__(attr, kwargs[attr])
            elif attr == "wavelength":
                assert isinstance(kwargs[attr], float), "'wavelength' should be a float value."
                self.__setattr__(attr, kwargs[attr])
            elif attr == "bandwidth":
                assert isinstance(kwargs[attr], float), "'bandwidth' should be a float value."
                self.__setattr__(attr, kwargs[attr])
            elif attr == "mean_photon_num":
                assert isinstance(kwargs[attr], float), f"'mean_photon_num' should be a float value."
                self.__setattr__(attr, kwargs[attr])
            else:
                raise TypeError(f"Setting {attr} is not allowed in {self.name}")

    def emit(self, dst: "Node", states: List, mean_photon_num_list=None) -> None:
        r"""Emit photons in specific states to the destination.

        Args:
            dst (Node): the destination node of the emitted photons
            states (List[numpy.ndarray]): expected quantum states to emit
            mean_photon_num_list (list): source intensity for each emitted states

        Important:
            The ``mean_photon_num_list`` is a list of the same length as ``states``, indicating the intensity of
            each emitted state. This will be used in protocols like decoy-state BB84.
            If ``mean_photon_num_list == None``, all states will be emitted with the same ``mean_photon_num``.
        """
        interval = round(1e12 / self.frequency)

        for i, state in enumerate(states):
            # Number of photons in a pulse (Poisson distribution)
            mean_photon_num_list = [self.mean_photon_num] * len(states) if mean_photon_num_list is None \
                else mean_photon_num_list
            photons_num = random.poisson(mean_photon_num_list[i])
            for _ in range(photons_num):
                wavelength = self.bandwidth * random.randn() + self.wavelength
                photon = Photon(wavelength=wavelength, state=MixedState(state, [0]), index=i)
                handler = EventHandler(self.owner, "send_quantum_msg", [dst, photon])
                self.scheduler.schedule_after(i * interval, handler)

    def print_parameters(self) -> None:
        r"""Print parameters of the photon source.
        """
        print(f"Details of Photon Source: {self.name}\n"
              f"frequency: {self.frequency} hertz\n"
              f"wavelength: {self.wavelength} nanometers\n"
              f"bandwidth: {self.bandwidth} nanometers\n"
              f"mean photon number: {self.mean_photon_num}")



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The backend provided by QCompute on Quantum Leaf.
"""

import QCompute

__all__ = [
    "DEFAULT_BACKEND",
    "DEFAULT_TOKEN",
    "set_qcompute_default_backend",
    "get_qcompute_default_backend",
    "set_qcompute_default_token",
    "get_qcompute_default_token",
    "run_circuit"
]


DEFAULT_BACKEND = 'local_baidu_sim2'
DEFAULT_TOKEN = None
QCompute.Define.Settings.outputInfo = False
QCompute.Define.Settings.drawCircuitControl = []
QCompute.Define.Settings.autoClearOutputDirAfterFetchMeasure = True
QCompute.Define.Settings.cloudTaskDoNotWriteFile = True


def set_qcompute_default_backend(backend: str) -> None:
    r"""Set the default backend for QCompute.

    Args:
        backend (str): the backend to set
    """
    assert isinstance(backend, str), "Should input the name of the backend in string."
    global DEFAULT_BACKEND
    DEFAULT_BACKEND = backend


def get_qcompute_default_backend() -> str:
    r"""Get the default backend of QCompute.

    Returns:
        str: current backend of QCompute
    """
    return DEFAULT_BACKEND


def set_qcompute_default_token(token: str) -> None:
    r"""Set the default token for QCompute backend.

    Args:
        token (str): your token for QCompute backend

    Warning:
        A valid token is required to use the QCompute cloud server. Please refer to the
        `website <https://quantum-hub.baidu.com/>`__ to get a valid token.
    """
    global DEFAULT_TOKEN
    DEFAULT_TOKEN = token


def get_qcompute_default_token() -> str:
    r"""Get the token you set.

    Returns:
        str: the current default token of QCompute
    """
    return DEFAULT_TOKEN


def run_circuit(circuit: "Circuit", shots: int, backend=None, token=None) -> dict:
    r"""Run a quantum circuit by QCompute backends.

    Args:
        circuit (Circuit): quantum circuit to run
        shots (int): number of sampling
        backend (Enum): specific QCompute backend
        token (str): your token for QCompute backend

    Raises:
        NotImplementedError: Some quantum gates are not supported in QCompute yet.

    Returns:
        dict: circuit sampling results

    Warning:
        Note that the labels of returning result is reversed from QCompute. That is, if the counting result
        from QCompute is given by ``{'01': 5, '00': 3}``, then our return gives ``{'10': 5, '00': 3}``.
    """
    single_qubit_gates = {
        'id': QCompute.ID, 's': QCompute.S, 't': QCompute.T,
        'h': QCompute.H, 'x': QCompute.X, 'y': QCompute.Y, 'z': QCompute.Z,
        'rx': QCompute.RX, 'ry': QCompute.RY, 'rz': QCompute.RZ, 'u3': QCompute.U,
        'm': QCompute.MeasureZ
    }
    multi_qubits_gates = {
        'ch': QCompute.CH, 'cx': QCompute.CX, 'cy': QCompute.CY, 'cz': QCompute.CZ,
        'crx': QCompute.CRX, 'cry': QCompute.CRY, 'crz': QCompute.CRZ, 'cu3': QCompute.CU,
        'swap': QCompute.SWAP
    }
    backend = QCompute.BackendName(get_qcompute_default_backend()) if backend is None else backend.value
    QCompute.Define.hubToken = get_qcompute_default_token() if token is None else token

    if backend != QCompute.BackendName.LocalBaiduSim2:
        assert QCompute.Define.hubToken is not None, \
            "A valid QCompute token is required to use the cloud backend. " \
            "Please set a QCompute token by calling `set_qcompute_default_token`."

    qcompute_env = QCompute.QEnv()
    qcompute_env.backend(backendName=backend)

    # Remap the indices to sequential integers starting from zero
    if circuit.width != max(circuit.occupied_indices) + 1:
        new_circuit = circuit.copy()
        new_circuit.remap_indices()
        circuit = new_circuit

    for gate in circuit.gate_history:
        name = gate['name']
        which_qubit = gate['which_qubit']

        if name in single_qubit_gates:
            gate_func = single_qubit_gates[name]
            if name in ['id', 's', 't', 'h', 'x', 'y', 'z']:
                gate_func(qcompute_env.Q[which_qubit[0]])
            elif name in ['rx', 'ry', 'rz']:
                gate_func(gate['angle'])(qcompute_env.Q[which_qubit[0]])
            elif name in ['u3']:
                gate_func(*gate['angles'])(qcompute_env.Q[which_qubit[0]])
            elif name in ['m']:
                gate_func([qcompute_env.Q[which_qubit[0]]], [which_qubit[0]])

        elif name in multi_qubits_gates:
            gate_func = multi_qubits_gates[name]
            if name in ['ch', 'cx', 'cy', 'cz', 'swap']:
                gate_func(qcompute_env.Q[which_qubit[0]], qcompute_env.Q[which_qubit[1]])
            elif name in ['crx', 'cry', 'crz']:
                gate_func(gate['angle'])(qcompute_env.Q[which_qubit[0]], qcompute_env.Q[which_qubit[1]])
            elif name in ['cu3']:
                gate_func(*gate['angles'])(qcompute_env.Q[which_qubit[0]], qcompute_env.Q[which_qubit[1]])

        else:
            raise NotImplementedError

    results = qcompute_env.commit(shots, fetchMeasure=True)
    if results['status'] == 'success':
        origin_counts = results['counts']
        counts = {}
        for key in origin_counts:
            key_reverse = key[::-1]  # reverse the bit string
            counts[key_reverse] = origin_counts[key]

        return counts

    else:
        raise InterruptedError("QCompute running failed.")



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for mobility of nodes.
"""

from abc import ABC, abstractmethod

__all__ = [
    "Mobility",
    "Track"
]


class Mobility:
    r"""Class for the mobility of a node.

    A mobile node moves on its preset track.

    Attributes:
        track (Track): track of the mobile node
    """

    def __init__(self):
        r"""Constructor for Mobility class.
        """
        self.track = None

    def set_track(self, track: "Track") -> None:
        r"""Set a track for the mobile node.

        Args:
            track (Track): track to set
        """
        self.track = track


class Track(ABC):
    r"""Class for the track of a mobile node.

    Attributes:
        ref_node (Node): reference node
        ref_time (int): time elapsed from the start of the simulation to the start of the movement
    """

    def __init__(self, ref_node: "Node", ref_time=0):
        r"""Constructor for Track class.

        Args:
            ref_node (Node): reference node
            ref_time (int): time elapsed from the start of the simulation to the start of the movement
        """
        self.ref_node = ref_node
        self.ref_time = ref_time

    @abstractmethod
    def time2distance(self, current_time: int) -> float:
        r"""Mapping the time to the distance from the mobile node to the reference node.

        Args:
            current_time (int): current time of the simulation

        Returns:
            float: the distance from the mobile node to the reference node
        """
        pass

    @abstractmethod
    def distance2loss(self, distance: float) -> float:
        r"""Mapping the distance to the communication loss.

        Args:
            distance (float): the distance from the mobile node to the reference node

        Returns:
            float: communication loss from the mobile node to the reference node
        """
        pass



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for quantum nodes.
"""

from typing import Optional

from qcompute_qnet.topology.node import Node
from qcompute_qnet.devices.register import QuantumRegister
from qcompute_qnet.models.qpu.message import QuantumMsg
from qcompute_qnet.protocols.protocol import ProtocolStack

__all__ = [
    "QuantumNode"
]


class QuantumNode(Node):
    r"""Node with a quantum register built-in.

    Attributes:
        qreg (QuantumRegister): quantum register installed
        protocol_type (type): protocol to install in the protocol stack
    """

    def __init__(self, name: str, qreg_size: int, protocol: Optional[type] = None, env: Optional["DESEnv"] = None):
        r"""Constructor for QuantumNode class.

        Args:
            name (str): name of the quantum node
            qreg_size (int): size of the local quantum register
            protocol (type, optional): protocol to install in the local protocol stack
            env (DESEnv, optional): discrete-event simulation environment
        """
        super().__init__(name, env)
        self.protocol_type = protocol

        self.qreg = QuantumRegister(name=f"{name}_qreg", size=qreg_size)  # create a quantum register
        self.install(self.qreg)
        ps = ProtocolStack(name=f"ps_{self.name}",
                           protocol=protocol(f"{protocol.__name__}_{self.name}") if protocol is not None else None)
        self.load_protocol(ps)

    @property
    def protocol(self) -> "Protocol":
        r"""Get the protocol from the protocol stack with the pre-specified type.

        Returns:
            Protocol: protocol with the pre-specified type
        """
        assert self.protocol_type is not None, f"Should set a protocol first."

        return self.protocol_stack.get_protocol(self.protocol_type)

    def send_quantum_msg(self, dst: "Node", qreg_address: int, priority=None) -> None:
        r"""Get the quantum state from the quantum register with a given address
        and send it as a quantum message to a given destination.

        Args:
            dst (Node): destination of the quantum message
            qreg_address (int): unit address in the quantum register
            priority (int, optional): priority of the quantum message
        """
        quantum_msg = QuantumMsg(self.qreg.get_qubit(qreg_address), self.qreg.circuit_index)
        super().send_quantum_msg(dst=dst, msg=quantum_msg, priority=priority)

    def receive_classical_msg(self, src: "Node", msg: "ClassicalMessage") -> None:
        r"""Receive a classical message from the classical channel.

        The source node is inferred by the receiver and is not part of the transmitted message.

        Args:
            src (Node): source of the classical message
            msg (ClassicalMessage): classical message to receive
        """
        assert self.protocol_type is not None, f"Should set a specific protocol type for the node first."

        self.protocol_stack.get_protocol(self.protocol_type).receive_classical_msg(msg)

    def receive_quantum_msg(self, src: "Node", msg: "QuantumMsg") -> None:
        r"""Receive a quantum message from the quantum channel.

        The source node is inferred by the receiver and is not part of the transmitted message.

        Args:
            src (Node): source of the quantum message
            msg (QuantumMessage): quantum message to receive
        """
        assert self.protocol_type is not None, f"Should set a specific protocol type for the node first."

        self.protocol_stack.get_protocol(self.protocol_type).receive_quantum_msg(msg, src=src)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for beam splitters.
"""

from typing import Tuple
from qcompute_qnet.core.des import Entity

__all__ = [
    "PolarizationBeamSplitter"
]


class PolarizationBeamSplitter(Entity):
    r"""Class for creating a polarization beam splitter.

    Attributes:
        clock (int): scheduled time of photon reception in picoseconds
        frequency (float): frequency of receiving photon pulse in hertz
        bases (List[numpy.ndarray]): measurement bases
        receivers (Tuple[Entity]): entities to receive transmitted photons
    """

    def __init__(self, name: str, env=None):
        r"""Constructor for polarization beam splitters.

        Args:
            name (str): name of the polarization beam splitter
            env (DESEnv): discrete-event simulation environment
        """
        super().__init__(name, env)
        self.clock = 0
        self.frequency = 0
        self.bases = []
        self.receivers = ()

    def init(self) -> None:
        r"""Polarization beam splitter initialization.
        """
        assert self.owner != self, f"The polarization beam splitter {self.name} has no owner!"

    def set(self, **kwargs) -> None:
        r"""Set given parameters.

        Args:
            **kwargs: keyword arguments to set
        """
        for attr in kwargs:
            if attr == "clock":
                assert isinstance(kwargs[attr], int), "'clock' should be an int value."
                self.__setattr__(attr, kwargs[attr])
            elif attr == "frequency":
                assert isinstance(kwargs[attr], float) or isinstance(kwargs[attr], int), \
                    "'frequency' should be a float or int value."
                self.__setattr__(attr, kwargs[attr])
            elif attr == "bases":
                assert isinstance(kwargs[attr], list), "'bases' should be a list."
                self.__setattr__(attr, kwargs[attr])
            elif attr == "receivers":
                assert isinstance(kwargs[attr], tuple) and len(kwargs[attr]) == 2, \
                    f"'receivers' should be a tuple of length 2."
                self.__setattr__(attr, kwargs[attr])
            else:
                raise TypeError(f"Setting {attr} is not allowed in {self.name}")

    def receive(self, photon: "Photon") -> None:
        r"""Receive an incoming photon.

        Args:
            photon (Photon): received photon
        """
        assert self.frequency != 0, "Should set a frequency first."
        outcome = photon.state.measure(self.bases[photon.index])
        self.receivers[outcome].receive(photon.index)

    def print_parameters(self) -> None:
        r"""Print parameters of the polarization beam splitter.
        """
        print("-" * 50)
        print(f"Details of Polarization BeamSplitter: {self.name}")
        print(f"frequency: {self.frequency}\n"
              f"receivers: {self.receivers}")
